import os
import json
import traceback
from .._data import _raise_fastai_import_error

HAS_TRANSFORMERS = True

try:
    import torch
    from transformers import AutoModelForSeq2SeqLM
    from ._arcgis_transformer import ArcGISTransformer
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_TRANSFORMERS = False
    class ArcGISTransformer:
        pass

backbone_models_map = {
    't5': ["t5-small","t5-base","t5-large","t5-3b","t5-11b"]+['See all T5 models at https://huggingface.co/models?filter=t5 '],
    'bart': ["facebook/bart-base","facebook/bart-large","facebook/bart-large-mnli","facebook/bart-large-cnn","facebook/bart-large-xsum","facebook/mbart-large-en-ro"]+['See all BART models at https://huggingface.co/models?filter=bart '],
    'marian': ['See all Marian models at https://huggingface.co/models?search=Helsinki-NLP ']
}

transformer_architectures = ['T5', 'Bart', 'Marian']

backbone_models_reverse_map = {x:key for key, val in backbone_models_map.items() for x in val}

transformer_seq_length = 512


class TransformerForSequenceToSequence(ArcGISTransformer):

    _supported_backbones = transformer_architectures

    def __init__(self, architecture, pretrained_model_name, config=None, pretrained_model_path=None,
                 seq_len=transformer_seq_length):   
        if not HAS_TRANSFORMERS:
            _raise_fastai_import_error(import_exception=import_exception)
        self.architecture = architecture
        self._seq_len = seq_len
        super().__init__(architecture, pretrained_model_name, config, pretrained_model_path, task='sequence_translation')
        self._seq_len = seq_len
        self._max_seq_len = None
        self.pretrained_model_name = pretrained_model_name

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % type(self).__name__

    def _process_config(self):
        """
        Process the model config to convert the class values to integers in the
        id2label attribute (which is a dictionary) of the model config
        """
        pass
        # self._config.id2label = {int(x): y for x, y in self._config.id2label.items()}

    def _set_max_seq_length(self):
        """
        Set the max-seq-len parameter to avoid out of memory issues for transformers
        like Bart, Longformer, XLNet
        """
        self._max_seq_len = min(self._tokenizer.model_max_length, self._seq_len)

    @classmethod
    def _available_backbone_models(cls, architecture):
        """
        Provides a list of available models for a given transformer architecture

        =====================   =================================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------------
        architecture            Required string. The transformer architecture for
                                which we wish to get the available models
        ---------------------   -------------------------------------------------
        return: tuple containing available models for the given architecture
        """
        if architecture.lower() in backbone_models_map:
            return backbone_models_map[architecture.lower()]
        else:
            return f"Error, wrong architecture name - {architecture} supplied. " \
                   f"PLease choose from {cls._supported_backbones}"

    def save(self, model_path):
        """
        Method to save the fine-tuned model to the disk

        =====================   =================================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------------
        model_path              Required string. The disk location where the
                                fine-tuned model has to be saved
        ---------------------   -------------------------------------------------
        """
        if not os.path.exists(model_path):
            os.mkdir(model_path)
        model_params = {"architecture": self._transformer_architecture,
                        "pretrained_model": self._transformer_pretrained_model_name}
        file_path = os.path.join(model_path, self._outfile)
        out_file = open(file_path, "w")
        json.dump(model_params, out_file)
        self._transformer.save_pretrained(model_path)

    @classmethod
    def load_pretrained_model(cls, path):
        """
        Method to load the fine-tuned model which was saved on the disk

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        path                    Required string. The disk location from
                                where the fine-tuned model has to be loaded
        ---------------------   -------------------------------------------
        """
        architecture_file_path = os.path.join(path, cls._outfile)
        if os.path.exists(architecture_file_path):
            with open(architecture_file_path) as f:
                data = json.load(f)
            architecture = data.get("architecture")
            pretrained_model_name = data.get("pretrained_model")
            obj = TransformerForSequenceToSequence(architecture, pretrained_model_name, pretrained_model_path=path)
            obj.init_model()
            return obj
        else:
            raise Exception(f"{cls._outfile} not present at {path}")

    def forward(self, input_ids, labels, **kwargs):
        """
        Return only the logits from the transfomer model

        =====================   ==============================================
        **Argument**            **Description**
        ---------------------   ----------------------------------------------
        input_ids               tensor object. tensor containing the token-ids
                                got by calling tokenizer.encode method to the
                                input text. works same for mini-batches
        ---------------------   ----------------------------------------------

        :return: the logits from the transformer model
        """
        if self.architecture in ['t5','bart','marian']:
            pad_id = self._tokenizer.pad_token_id
            attention_mask = (input_ids!=pad_id).int()
            output_dict = self._transformer(input_ids=input_ids, attention_mask=attention_mask, labels = labels, return_dict = True)
            logits = output_dict.get('logits')  
            loss = output_dict.get('loss')
            return loss, logits

    def _load_transformer(self):
        """
        Method to load the transformer model. For this class we want to load the
        transformer model for performing Sequence Classification tasks
        """
        if self._pretrained_model_path and os.path.exists(self._pretrained_model_path):
            base_path, extension = os.path.splitext(self._pretrained_model_path)
            path = base_path + ".pth"
            if extension == ".emd" and os.path.exists(path):
                self._transformer = AutoModelForSeq2SeqLM.from_pretrained(path, config=self._config)
            else:
                self._transformer = AutoModelForSeq2SeqLM.from_pretrained(self._pretrained_model_path)
        else:
            self._transformer = AutoModelForSeq2SeqLM.\
                from_pretrained(self._transformer_pretrained_model_name, config=self._config)



"""

These tools are used for data enrichment using geoanalytics

"""
import json as _json
import logging as _logging
import arcgis as _arcgis
from arcgis.features import FeatureSet as _FeatureSet
from arcgis.geoprocessing._support import _execute_gp_tool
from arcgis.geoanalytics._util import (_id_generator,
                                       _feature_input,
                                       _set_context,
                                       _create_output_service,
                                       GAJob,
                                       _prevent_bds_item)

_log = _logging.getLogger(__name__)

_use_async = True


def calculate_motion_statistics(input_layer,
                                track_fields,
                                motion_statistics="All",
                                track_history_window=3,
                                idle_tol_dist=None,
                                idle_tol_unit=None,
                                idle_time_tol=None,
                                idle_time_tol_unit=None,
                                time_boundary_split=None,
                                split_unit=None,
                                time_bound_ref=None,
                                dist_method="Geodesic",
                                distance_unit="Meters",
                                duration_unit="Seconds",
                                speed_unit="MetersPerSecond",
                                accel_unit="MetersPerSecondSquared",
                                elev_unit="meters",
                                output_name=None,
                                gis=None,
                                context=None,
                                future=False):
    """
    The Calculate Motion Statistics task calculates motion statistics and
    descriptors for time-enabled points that represent one or more moving
    entities. Points are grouped together into tracks representing each
    entity using a unique identifier. Motion statistics are calculated at
    each point using one or more points in the track history. Calculations
    include summaries of distance traveled, duration, elevation, speed,
    acceleration, bearing, and idle status. The result is a new point layer
    enriched with the requested statistics.

    For example, a city is monitoring snowplow operations and wants to
    better understand vehicle movement. The Calculate Motion Statistics
    tool can be used to determine idle locations and time spent idling,
    average and maximum speeds over time, total distance covered, and other
    statistics.

    .. note::
        Only available at ArcGIS Enterprise 10.9 and later.

    ======================  ===============================================================
    **Argument**            **Description**
    ----------------------  ---------------------------------------------------------------
    input_layer             Required layer. The time-enabled point features that will be
                            grouped into tracks and analyzed. The input layer must be of
                            time type instant. See :ref:`Feature Input<gaxFeatureInput>`.
    ----------------------  ---------------------------------------------------------------
    track_fields            Required String. The fields used to identify distinct tracks.
                            There can be multiple trackFields in seperated by commas.
    ----------------------  ---------------------------------------------------------------
    motion_statistics       Optional String. The type of motion statistics to calulcated.
                            The allowed values are: `distance`, `speed`, `acceleration`,
                            `duration` or `elevation`, `slope`, `idle`, `bearing`, or `all` (default).

    ----------------------  ---------------------------------------------------------------
    track_history_window    Optional Integer. The number of observations (including the
                            current observation) that will be used when calculating summary
                            statistics that are not instantaneous. This includes minimum,
                            maximum, average, and total statistics. The default track history
                            window is **3**, which means that at each point in a track
                            summary statistic will be calculated using the current
                            observation and the previous three observations. This parameter
                            does not affect instantaneous statistics or idle classification.
    ----------------------  ---------------------------------------------------------------
    idle_tol_dist           Optional Float. Used along with `idle_time_tol` to decide if an
                            entity is idling. An entity is idling when it hasn't moved more
                            than this distance in at least the amount of time specified by
                            `idle_time_tol`. The units of the time values are supplied by
                            the `idle_tol_dist` parameter.

                            This value is only used for statistics in the Idle group.
    ----------------------  ---------------------------------------------------------------
    idle_tol_unit           Optional String. The unit of distance for idle distance tolerance.
    ----------------------  ---------------------------------------------------------------
    idle_time_tol           Optional Float. The lead amount of time used to determine if someone is idling.
    ----------------------  ---------------------------------------------------------------
    idle_time_tol_unit      Optional String. The time tolerance unit for idling.
    ----------------------  ---------------------------------------------------------------
    time_boundary_split     Optional Float. A time boundary allows your to analyze values within a defined time span.
    ----------------------  ---------------------------------------------------------------
    split_unit              Optional String. The unit of time represented in the `time_boundary_split`.
    ----------------------  ---------------------------------------------------------------
    time_bound_ref          Optional Datetime. A date that specifies the reference time to
                            align the time boundary to, represented in milliseconds from epoch.
    ----------------------  ---------------------------------------------------------------
    dist_method             Optional String. The method used to calculate distances between
                            track observations. There are two methods to choose from: `Planar`
                            and `Geodesic`. The `Planar` method measures distances using an
                            Euclidean plane and will not calculate statistics across the
                            date line. When the `Geodesic` method is used to calculate
                            distance and the spatial reference can be panned, calculations
                            will cross the date line when appropriate. If the spatial
                            reference cannot be panned, calculations will be limited to the
                            coordinate system extent and may not wrap.
    ----------------------  ---------------------------------------------------------------
    distance_unit           Optional String. The units for all results in the Distance
                            motion statistics group.
                            Values: Meters (default) | Kilometers | Feet | Miles | NauticalMiles | Yards
    ----------------------  ---------------------------------------------------------------
    duration_unit           Optional String. The units for all results in the Duration motion statistics group.

                            Values: Milliseconds | Seconds (default) | Minutes | Hours | Days | Weeks| Months | Years
    ----------------------  ---------------------------------------------------------------
    speed_unit              Optional String. The units for all results in the Speed motion statistics group.

                            Values: MetersPerSecond (default) | KilometersPerHour | FeetPerSecond | MilesPerHour | NauticalMilesPerHour
    ----------------------  ---------------------------------------------------------------
    accel_unit              Optional String. The units for all results in the Acceleration motion statistics group.

                            Values: MetersPerSecondSquared (default) | FeetPerSecondSquared
    ----------------------  ---------------------------------------------------------------
    elev_unit               Optional String. The units for all results in the Elevation motion statistics group.

                            Values: Meters (default) | Kilometers | Feet | Miles | NauticalMiles | Yards
    ----------------------  ---------------------------------------------------------------
    output_name             optional string. The task will create a feature service of the
                            results. You define the name of the service.
    ----------------------  ---------------------------------------------------------------
    gis                     optional GIS. The GIS object where the analysis will take place.
    ----------------------  ---------------------------------------------------------------
    context                 Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are five settings:

                            #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                            #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                            #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                            #. Data store (``dataStore``) - Results will be saved to the specified data store. For ArcGIS Enterprise, the default is the spatiotemporal big data store.
                            #. Default aggregation styles (``defaultAggregationStyles``) - If set to 'True', results will have square, hexagon, and triangle aggregation styles enabled on results map services.
    ----------------------  ---------------------------------------------------------------
    future                  optional boolean. If 'True', a GPJob is returned instead of
                            results. The GPJob can be queried on the status of the execution.

                            The default value is 'False'.
    ======================  ===============================================================

    :returns: result_layer : Output Features as feature layer item.

    """
    kwargs = locals()
    input_layer = _prevent_bds_item(input_layer)
    tool_name = "CalculateMotionStatistics"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json",
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Calculate_Motion_Stats_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')
    if context is not None:
        output_datastore = context.get('dataStore', None)
    else:
        output_datastore = None
    output_service = _create_output_service(gis, output_name, output_service_name, 'Calculate Motion Statistics', output_datastore=output_datastore)

    if output_service:
        params['output_name'] = _json.dumps({
            "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
            "itemProperties": {"itemId" : output_service.itemid}})
    else:
        params['output_name'] = output_service_name
        output_service = f"Results were written to: '{params['context']['dataStore']}' with the name: '{output_service_name}'"

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "track_fields": (str, "trackFields"),
        "track_history_window" : (int, "trackHistoryWindow"),
        "motion_statistics" : (str, "motionStatistics"),
        "idle_tol_dist" : (str,"idleDistanceTolerance"),
        "idle_tol_unit" : (str, "idleDistanceToleranceUnit"),
        "idle_time_tol" : (float, "idleTimeTolerance"),
        "idle_time_tol_unit" : (str, "idleTimeToleranceUnit"),
        "time_boundary_split" : (int, "timeBoundarySplit"),
        "split_unit": (str, "timeBoundarySplitUnit"),
        "time_bound_ref" : (int, "timeBoundaryReference"),
        "dist_method" : (str, "distanceMethod"),
        "distance_unit" : (str, "distanceUnit"),
        "duration_unit" : (str, "durationUnit"),
        "speed_unit" : (str, "speedUnit"),
        "accel_unit" : (str, "accelerationUnit"),
        "elev_unit" : (str, "elevationUnit"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "output"),
    }

    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        if future:
            gpjob = _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise



def enrich_from_grid(input_layer,
                     grid_layer,
                     enrichment_attributes=None,
                     output_name=None,
                     gis=None,
                     context=None,
                     future=False):
    """
    .. image:: _static/images/enrich_from_grid/enrich_from_grid.png

    The Enrich From Multi-Variable Grid task joins attributes from a multivariable grid to a point layer.
    The multivariable grid must be created using the ``build_multivariable_grid`` task. Metadata from the
    multivariable grid is used to efficiently enrich the input point features, making it faster than the
    Join Features task. Attributes in the multivariable grid are joined to the input point features when
    the features intersect the grid.

    The attributes in the multivariable grid can be used as explanatory variables when modeling spatial
    relationships with your input point features, and this task allows you to join those attributes to
    the point features quickly.

    .. note::
        Only available at ArcGIS Enterprise 10.7 and later.

    ======================  ===============================================================
    **Argument**            **Description**
    ----------------------  ---------------------------------------------------------------
    input_layer             Required layer. The point features that will be enriched
                            by the multi-variable grid. See :ref:`Feature Input<gaxFeatureInput>`.
    ----------------------  ---------------------------------------------------------------
    grid_layer              Required layer. The multivariable grid layer created using the Build Multi-Variable Grid task.
                            See :ref:`Feature Input<gaxFeatureInput>`.
    ----------------------  ---------------------------------------------------------------
    enrichment_attributes   optional string. A list of fields in the multi-variable grid
                            that will be joined to the input point features. If the
                            attributes are not provided, all fields in the multi-variable
                            grid will be joined to the input point features.
    ----------------------  ---------------------------------------------------------------
    output_name             optional string. The task will create a feature service of the
                            results. You define the name of the service.
    ----------------------  ---------------------------------------------------------------
    gis                     optional GIS. The GIS object where the analysis will take place.
    ----------------------  ---------------------------------------------------------------
    context                 Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are five settings:

                            #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                            #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                            #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                            #. Data store (``dataStore``) - Results will be saved to the specified data store. For ArcGIS Enterprise, the default is the spatiotemporal big data store.
                            #. Default aggregation styles (``defaultAggregationStyles``) - If set to 'True', results will have square, hexagon, and triangle aggregation styles enabled on results map services.
    ----------------------  ---------------------------------------------------------------
    future                  optional boolean. If 'True', a GPJob is returned instead of
                            results. The GPJob can be queried on the status of the execution.

                            The default value is 'False'.
    ======================  ===============================================================

    :returns: result_layer : Output Features as feature layer item.

    .. code-block:: python

            # Usage Example: To enrich a layer of crime data with a multivariable grid containing demographic information.

            enrich_result = enrich_from_grid(input_layer=crime_lyr,
                                             grid_layer=mvg_layer,
                                             output_name="chicago_crimes_enriched")


    """
    kwargs = locals()
    input_layer = _prevent_bds_item(input_layer)
    tool_name = "EnrichFromMultiVariableGrid"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json",
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Enrich_Grid_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')
    if context is not None:
        output_datastore = context.get('dataStore', None)
    else:
        output_datastore = None
    output_service = _create_output_service(gis, output_name, output_service_name, 'Enrich Grid Layers', output_datastore=output_datastore)

    if output_service:
        params['output_name'] = _json.dumps({
            "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
            "itemProperties": {"itemId" : output_service.itemid}})
    else:
        params['output_name'] = output_service_name
        output_service = f"Results were written to: '{params['context']['dataStore']}' with the name: '{output_service_name}'"

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputFeatures"),
        "grid_layer" : (_FeatureSet, "gridLayer"),
        "enrichment_attributes" : (str, "enrichAttributes"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "output"),
    }

    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        if future:
            gpjob = _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise




try:
    import os, sys, json, importlib
    import numpy as np
    import torch
    import torch.nn as nn
    from fastai.core import split_kwargs_by_func
    import math
    from . import util
    from .util import normalize_batch
    from pathlib import Path
    HAS_TORCH = True
except Exception as e:
    HAS_TORCH = False
    
import arcgis
from arcgis.learn import ModelExtension

try:
    import arcpy
except:
    pass

def normalize_batch_imagenetstats(batch):

    imagenet_stats = [[0.485, 0.456, 0.406], [0.229, 0.224, 0.225]]
    mean = 255* np.array(imagenet_stats[0], dtype=np.float32)
    std  = 255* np.array(imagenet_stats[1], dtype=np.float32)

    return (batch-mean)/ std

def calculate_rectangle_size_from_batch_size(batch_size):
    '''
    calculate number of rows and cols to composite a rectangle given a batch size
    :param batch_size:
    :return: number of cols and number of rows
    '''
    rectangle_height = int(math.sqrt(batch_size) + 0.5)
    rectangle_width = int(batch_size / rectangle_height)

    if rectangle_height * rectangle_width > batch_size:
        if rectangle_height >= rectangle_width:
            rectangle_height = rectangle_height - 1
        else:
            rectangle_width = rectangle_width - 1

    if (rectangle_height + 1) * rectangle_width <= batch_size:
        rectangle_height = rectangle_height + 1
    if (rectangle_width + 1) * rectangle_height <= batch_size:
        rectangle_width = rectangle_width + 1

    # swap col and row to make a horizontal rect
    if rectangle_height > rectangle_width:
        rectangle_height, rectangle_width = rectangle_width, rectangle_height

    if rectangle_height * rectangle_width != batch_size:
        return batch_size, 1

    return rectangle_height, rectangle_width

def convert_bounding_boxes_to_coord_list(bounding_boxes):
    '''
    convert bounding box numpy array to python list of point arrays
    :param bounding_boxes: numpy array of shape [n, 4]
    :return: python array of point numpy arrays, each point array is in shape [4,2]
    '''
    num_bounding_boxes = bounding_boxes.shape[0]
    bounding_box_coord_list = []
    for i in range(num_bounding_boxes):
        coord_array = np.empty(shape=(4, 2), dtype=np.float)
        coord_array[0][0] = bounding_boxes[i][0]
        coord_array[0][1] = bounding_boxes[i][1]

        coord_array[1][0] = bounding_boxes[i][0]
        coord_array[1][1] = bounding_boxes[i][3]

        coord_array[2][0] = bounding_boxes[i][2]
        coord_array[2][1] = bounding_boxes[i][3]

        coord_array[3][0] = bounding_boxes[i][2]
        coord_array[3][1] = bounding_boxes[i][1]

        bounding_box_coord_list.append(coord_array)

    return bounding_box_coord_list


def get_tile_size(model_height, model_width, padding, batch_height, batch_width):
    '''
    Calculate request tile size given model and batch dimensions
    :param model_height:
    :param model_width:
    :param padding:
    :param batch_width:
    :param batch_height:
    :return: tile height and tile width
    '''
    tile_height = (model_height - 2 * padding) * batch_height
    tile_width = (model_width - 2 * padding) * batch_width

    return tile_height, tile_width

def tile_to_batch(pixel_block, model_height, model_width, padding, fixed_tile_size=True, **kwargs):
    inner_width = model_width - 2 * padding
    inner_height = model_height - 2 * padding

    band_count, pb_height, pb_width = pixel_block.shape
    pixel_type = pixel_block.dtype

    if fixed_tile_size is True:
        batch_height = kwargs['batch_height']
        batch_width = kwargs['batch_width']
    else:
        batch_height = math.ceil((pb_height - 2 * padding) / inner_height)
        batch_width = math.ceil((pb_width - 2 * padding) / inner_width)

    batch = np.zeros(shape=(batch_width * batch_height, band_count, model_height, model_width), dtype=pixel_type)
    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        # pixel block might not be the shape (band_count, model_height, model_width)
        sub_pixel_block = pixel_block[:, y * inner_height: y * inner_height + model_height,
                    x * inner_width: x * inner_width + model_width]
        sub_pixel_block_shape = sub_pixel_block.shape
        batch[b, :, :sub_pixel_block_shape[1], :sub_pixel_block_shape[2]] = sub_pixel_block

    return batch, batch_height, batch_width


def batch_to_tile(batch, batch_height, batch_width):
    batch_size, bands, inner_height, inner_width = batch.shape
    tile = np.zeros(shape=(bands, inner_height * batch_height, inner_width * batch_width), dtype=batch.dtype)

    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        tile[:, y * inner_height: (y+1) * inner_height, x * inner_width:(x+1) * inner_width] = batch[b]

    return tile

def remove_bounding_boxes_in_padding(bounding_boxes, scores, classes, image_height, image_width, padding,
                                     batch_height=1, batch_width=1):
    '''

    :param bounding_boxes: the batch of bounding boxes, shape=[B,N,4]
    :param scores: the batch of box scores, shape=[B,N]
    :param classes: the batch of labels, shape=[B,N]
    :param image_height: model height
    :param image_width: model width
    :param padding:
    :param batch_height:
    :param batch_width:
    :return:
    '''
    keep_indices = np.where((bounding_boxes[:,:,0] < image_height-padding) &
              (bounding_boxes[:,:,1] < image_width-padding) &
              (bounding_boxes[:,:,2] > padding) &
              (bounding_boxes[:,:,3] > padding))

    inner_width = image_width - 2 * padding
    inner_height = image_height - 2 * padding

    # convert coordinates in the batch to super tile and then filter by the keep_indices
    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        bounding_boxes[b, :, [0, 2]] = bounding_boxes[b, :, [0, 2]] + y * inner_height
        bounding_boxes[b, :, [1, 3]] = bounding_boxes[b, :, [1, 3]] + x * inner_width

    bounding_boxes = bounding_boxes[keep_indices]
    scores = scores[keep_indices]
    classes = classes[keep_indices]

    return bounding_boxes, scores, classes

class ChildObjectDetector:

    def initialize(self, model, model_as_file):

        if not HAS_TORCH:
            raise Exception('PyTorch is not installed. Install it using conda install -c pytorch pytorch torchvision')

        if arcpy.env.processorType == "GPU" and torch.cuda.is_available():
            self.device = torch.device('cuda')
            arcgis.env._processorType = "GPU"
        else:
            self.device = torch.device('cpu')
            arcgis.env._processorType = "CPU"

        if model_as_file:
            with open(model, 'r') as f:
                self.json_info = json.load(f)
        else:
            self.json_info = json.load(model)

        model_path = self.json_info['ModelFile']
        if model_as_file and not os.path.isabs(model_path):
            model_path = os.path.abspath(os.path.join(os.path.dirname(model), model_path))

        self.json_emd_file = Path(model).parent
        self.model_extension = ModelExtension.from_model(emd_path=model)
        self.model = self.model_extension.learn.model.to(self.device)
        self.model.eval()
        
    def getParameterInfo(self, required_parameters):
        required_parameters.extend(
            [
                {
                    'name': 'padding',
                    'dataType': 'numeric',
                    'value': self.json_info['ImageHeight'] // 4,
                    'required': False,
                    'displayName': 'Padding',
                    'description': 'Padding'
                },
                {
                    'name': 'threshold',
                    'dataType': 'numeric',
                    'value': 0.5,
                    'required': False,
                    'displayName': 'Confidence Score Threshold [0.0, 1.0]',
                    'description': 'Confidence score threshold value [0.0, 1.0]'
                },
                {
                    'name': 'nms_overlap',
                    'dataType': 'numeric',
                    'value': 0.1,
                    'required': False,
                    'displayName': 'NMS Overlap',
                    'description': 'Maximum allowed overlap within each chip'
                },
                {
                    'name': 'batch_size',
                    'dataType': 'numeric',
                    'required': False,
                    'value': 64,
                    'displayName': 'Batch Size',
                    'description': 'Batch Size'
                },
                {
                    'name': 'exclude_pad_detections',
                    'dataType': 'string',
                    'required': False,
                    'domain': ('True', 'False'),
                    'value': 'True',
                    'displayName': 'Filter Outer Padding Detections',
                    'description': 'Filter detections which are outside the specified padding'
                }
                
            ]
        )
        return required_parameters

    def getConfiguration(self, **scalars):        
        self.padding = int(scalars.get('padding', self.json_info['ImageHeight'] // 4)) ## Default padding Imageheight//4.
        self.nms_overlap = float(scalars.get('nms_overlap', 0.1))  ## Default 0.1 NMS Overlap.
        self.thres = float(scalars.get('threshold', 0.5)) ## Default 0.5 threshold.
        self.batch_size = int(math.sqrt(int(scalars.get('batch_size', 64)))) ** 2  ## Default 64 batch_size
        self.filter_outer_padding_detections = scalars.get('exclude_pad_detections', 'True').lower() in ['true', '1', 't', 'y', 'yes'] ## Default value True 

        self.rectangle_height, self.rectangle_width = calculate_rectangle_size_from_batch_size(self.batch_size)
        ty, tx = get_tile_size(self.json_info['ImageHeight'], self.json_info['ImageWidth'],
                                         self.padding, self.rectangle_height, self.rectangle_width)

        return {
            'extractBands': tuple(self.json_info['ExtractBands']),
            'padding': self.padding,
            'threshold': self.thres,
            'nms_overlap': self.nms_overlap,
            'tx': tx,
            'ty': ty,
            'fixedTileSize': 1
        }

    def vectorize(self, **pixelBlocks): # 8 x 3 x 224 x 224

        input_image = pixelBlocks['raster_pixels'].astype(np.float32)
        batch, batch_height, batch_width = \
            tile_to_batch(input_image,
                        self.json_info['ImageHeight'],
                        self.json_info['ImageWidth'],
                        self.padding,
                        fixed_tile_size=True,
                        batch_height=self.rectangle_height,
                        batch_width=self.rectangle_width)      


        if "NormalizationStats" in self.json_info:
            img_normed = normalize_batch(batch, self.json_info)
        else:
            img_normed = normalize_batch_imagenetstats(batch.transpose(0, 2, 3, 1)).transpose(0, 3, 1, 2)

        bounding_boxes, scores, classes = detect_object(self.model_extension.model_conf,
                                                        self.model,
                                                        img_normed,
                                                        self.device,
                                                        batch_size=self.batch_size,
                                                        model_info=self.json_info,
                                                        emd_path=self.json_emd_file,
                                                        nms_overlap=self.nms_overlap,
                                                        thresh=self.thres)
        
        return convert_bounding_boxes_to_coord_list(bounding_boxes), scores, classes


def detect_object(model_configuration, model, images, device, batch_size, model_info, emd_path, **kwargs):

    tile_height, tile_width = images.shape[2], images.shape[3]
    side = math.sqrt(batch_size)
    thres = kwargs.get('thresh', 0.5)
    nms_overlap = kwargs.get('nms_overlap', 0.1)

    if "NormalizationStats" in model_info:
        transform_kwargs, kwargs = split_kwargs_by_func(kwargs, model_configuration.transform_input_multispectral)
        batch_input = model_configuration.transform_input_multispectral(torch.tensor(images).to(device).float(),
                                                                         **transform_kwargs)
    else:
        transform_kwargs, kwargs = split_kwargs_by_func(kwargs, model_configuration.transform_input)
        batch_input = model_configuration.transform_input(torch.tensor(images).to(device).float(), **transform_kwargs)

    pred_batch = model(batch_input)

    preds = model_configuration.post_process(pred_batch, nms_overlap, thres, tile_height, device)

    num_boxes = 0
    for batch_idx in range(batch_size):
        num_boxes = num_boxes + preds[batch_idx][0].size(0)
    
    bounding_boxes = np.empty(shape=(num_boxes, 4), dtype=np.float)
    scores = np.empty(shape=(num_boxes), dtype=np.float)
    classes = np.empty(shape=(num_boxes), dtype=np.uint8)

    idx = 0
    for batch_idx in range(batch_size):
        i, j = batch_idx//side, batch_idx%side

        for bbox, label, score in zip(preds[batch_idx][0], preds[batch_idx][1], preds[batch_idx][2]):
            bbox = (bbox+1)/2
            bounding_boxes[idx, 0] = (bbox.data[0] + i)*tile_height
            bounding_boxes[idx, 1] = (bbox.data[1] + j)*tile_width
            bounding_boxes[idx, 2] = (bbox.data[2] + i)*tile_height
            bounding_boxes[idx, 3] = (bbox.data[3] + j)*tile_width
            scores[idx] = score.data*100
            classes[idx] = label.data
            idx = idx+1

    return bounding_boxes, scores, classes


class ChildImageClassifier:

    def initialize(self, model, model_as_file):

        if not HAS_TORCH:
            raise Exception('PyTorch is not installed. Install it using conda install -c pytorch pytorch torchvision')

        if arcpy.env.processorType == "GPU" and torch.cuda.is_available():
            self.device = torch.device('cuda')
            arcgis.env._processorType = "GPU"
        else:
            self.device = torch.device('cpu')
            arcgis.env._processorType = "CPU"

        if model_as_file:
            with open(model, 'r') as f:
                self.json_info = json.load(f)
        else:
            self.json_info = json.load(model)

        model_path = self.json_info['ModelFile']
        if model_as_file and not os.path.isabs(model_path):
            model_path = os.path.abspath(os.path.join(os.path.dirname(model), model_path))

        self.json_emd_file = Path(model).parent
        self.model_extension = ModelExtension.from_model(emd_path=model)
        self.model = self.model_extension.learn.model.to(self.device)
        self.model.eval()
   
    def getParameterInfo(self, required_parameters):

        required_parameters.extend(
            [
                {
                    'name': 'padding',
                    'dataType': 'numeric',
                    'value': int(self.json_info['ImageHeight'])//4,
                    'required': False,
                    'displayName': 'Padding',
                    'description': 'Padding'
                },
                {
                    'name': 'batch_size',
                    'dataType': 'numeric',
                    'required': False,
                    'value': 4,
                    'displayName': 'Batch Size',
                    'description': 'Batch Size'
                }
            ]
        )
        if self.json_info['IsEdgeDetection']:
            required_parameters.extend(
                [
                    {
                        'name': 'thinning',
                        'dataType': 'string',
                        'value': 'False',
                        'required': False,
                        'displayName': 'thinning',
                        'description': 'If True, edges will be thined to one pixel wide'
                    }
                ]
            )
        else:
            required_parameters.extend(
                [
                    {
                    'name': 'predict_background',
                    'dataType': 'string',
                    'required': False,
                    'value': 'True',
                    'displayName': 'Predict Background',
                    'description': 'If False, will never predict the background/NoData Class.'
                    }
                ]
            )

        
        return required_parameters

    def getConfiguration(self, **scalars):
        self.padding = int(scalars.get('padding', self.json_info['ImageHeight'] // 4)) ## Default padding Imageheight//4.
        self.batch_size = int(math.sqrt(int(scalars.get('batch_size', 4)))) ** 2  ## Default 4 batch_size
        self.predict_background = scalars.get('predict_background', 'true').lower() in ['true', '1', 't', 'y', 'yes']  ## Default value True
        if self.json_info['IsEdgeDetection']:
            self.thinning = scalars.get('thinning', 'true').lower() in ['true', '1', 't', 'y', 'yes']
        else:
            self.thinning = None

        self.rectangle_height, self.rectangle_width = calculate_rectangle_size_from_batch_size(self.batch_size)
        ty, tx = get_tile_size(self.json_info['ImageHeight'], self.json_info['ImageWidth'],
                                         self.padding, self.rectangle_height, self.rectangle_width)

        return {
            'extractBands': tuple(self.json_info['ExtractBands']),
            'padding': self.padding,
            'tx': tx,
            'ty': ty,
            'fixedTileSize': 1
        }

    def updatePixels(self, tlc, shape, props, **pixelBlocks): # 8 x 224 x 224 x 3
        input_image = pixelBlocks['raster_pixels'].astype(np.float32)
        batch, batch_height, batch_width = \
            tile_to_batch(input_image,
                        self.json_info['ImageHeight'],
                        self.json_info['ImageWidth'],
                        self.padding,
                        fixed_tile_size=True,
                        batch_height=self.rectangle_height,
                        batch_width=self.rectangle_width)
        
        if "NormalizationStats" in self.json_info:
            img_normed = normalize_batch(batch, self.json_info)
        else:
            img_normed = normalize_batch_imagenetstats(batch.transpose(0, 2, 3, 1)).transpose(0, 3, 1, 2)

        semantic_predictions = classify_image(self.model_extension.model_conf,
                                            self.model,
                                            img_normed,
                                            self.device,
                                            predict_bg=self.predict_background,
                                            model_info=self.json_info,
                                            emd_path=self.json_emd_file,
                                            thinning=self.thinning)

        semantic_predictions = batch_to_tile(semantic_predictions.cpu().numpy(), batch_height, batch_width)
        return semantic_predictions

def classify_image(model_configuration, model, images, device, predict_bg, model_info, emd_path, thinning):

    if "NormalizationStats" in model_info:
        batch_input = model_configuration.transform_input_multispectral(torch.tensor(images).to(device).float())
    else:
        batch_input = model_configuration.transform_input(torch.tensor(images).to(device).float())

    pred_batch = model(batch_input)

    if thinning == None:
        preds = model_configuration.post_process(pred_batch)
        return torch.stack(preds)
    else:
        preds = model_configuration.post_process(pred_batch, thinning=thinning)
        if thinning:
            return torch.stack(preds)
        else:
            return preds



import torch
import torch.nn as nn


class _TSLSTM(nn.Module):
    def __init__(self, input_size=1, output_size=1, device=None, hidden_layer_size=100):
        super().__init__()
        self.hidden_layer_size = hidden_layer_size

        self.lstm = nn.LSTM(input_size, hidden_layer_size)

        self.linear = nn.Linear(hidden_layer_size, output_size)
        self.device = device
        self.hidden_cell = (torch.zeros(1, 1, self.hidden_layer_size).to(device),
                            torch.zeros(1, 1, self.hidden_layer_size).to(device))

    def forward(self, input_seq):
        outputs = []
        for batch in input_seq:
            batch = batch.squeeze()
            self.hidden_cell = (torch.zeros(1, 1, self.hidden_layer_size).to(self.device),
                                torch.zeros(1, 1, self.hidden_layer_size).to(self.device))

            lstm_out, self.hidden_cell = self.lstm(batch.view(len(batch), 1, -1).to(self.device), self.hidden_cell)
            predictions = self.linear(lstm_out.view(len(batch), -1))
            outputs += [predictions[-1]]

        outputs = torch.stack(outputs, 1)

        return outputs.to(self.device)


import os
import json
import random
import logging
import tempfile
import datetime
import traceback
from pathlib import Path
from .text_data import TextDataObject
try:
    import spacy
    from spacy.gold import offsets_from_biluo_tags as _offsets_from_biluo_tags
    from spacy.gold import iob_to_biluo as _iob_to_biluo
    import pandas as pd
    import numpy as np
    HAS_SPACY = True
except Exception as e:
    spacy_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_SPACY = False


__all__=["_from_iob_tags", "_from_json", "_NERData", "even_mults", "_timelapsed"]


def _raise_spacy_import_error():
    error_message = (f"{spacy_exception}\n\n\n"
                     "This module requires spacy version 2.1.8 or above and fastprogress." 
                     "Install it using 'pip install spacy==2.1.8 fastprogress pandas'")
    raise Exception(error_message)


def even_mults(start:float, stop:float, n:int): #Taken from FastAI(https://github.com/fastai/fastai/blob/master/fastai/core.py#L150)
    "Build log-stepped array from `start` to `stop` in `n` steps."
    mult = stop/start
    step = mult**(1/(n-1))
    return np.array([start*(step**i) for i in range(n)])


def _timelapsed(t_start):
    '''returns timedelta in hh:mm:ss format'''
    b = datetime.datetime.now() - t_start
    h,r = divmod(b.seconds, 3600)
    m,s = divmod(r, 60)
    return ("%02d:%02d:%02d"%(h,m,s))


def _from_iob_tags(tokens_collection, tags_collection):
    """
    Converts training data from ``IOB`` format to spacy offsets.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    tokens_collection       Required [list]. List of token lists
                            Example: [[This,is,a,test],[This,is,a,test1]]
    ---------------------   -------------------------------------------
    tags_collection         Required [list]. List of tag lists
                            Example: [[B-tagname,O,O,O],[B-tagname,I-tagname,O,O]]
    =====================   ===========================================
    """

    train_data = []
    nlp=spacy.blank('en')
    for tags, tokens in zip(tags_collection, tokens_collection):
        
        try:
            tags = _iob_to_biluo(tags)

            doc = spacy.tokens.doc.Doc(
            nlp.vocab, words = tokens, spaces = [True]*(len(tokens)-1)+[False])
            # run the standard pipeline against it
            for name, proc in nlp.pipeline:
                doc = proc(doc)
            
            text=' '.join(tokens)
            tags = _offsets_from_biluo_tags(doc, tags)
            train_data.append((text,{'entities':tags}))
        except:
            pass
        

    return train_data


def _from_json(path, text_key='text', offset_key='labels', encoding="UTF-8"):
    """
    Converts training data from JSON format to spacy offsets.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    text_key                Optional:str='text. Json key under which text is available
    ---------------------   -------------------------------------------
    offset_key              Optional:str='labels. Json key under which offsets are available
    =====================   ===========================================
    json-schema:
    ----------
    {"id": 1, "text": "EU rejects ...", "labels": [[0,2,"ORG"], [11,17, "MISC"], [34,41,"ORG"]]}
    {"id": 2, "text": "Peter Blackburn", "labels": [[0, 15, "PERSON"]]}
    {"id": 3, "text": "President Obama", "labels": [[10, 15, "PERSON"]]}
    ----------
    returns: A list that can be consumed by ner_databunch.
    """
    
    train_data = []
    with open(path,'r', encoding=encoding) as f:
        data_list = f.readlines()
    for i, item in enumerate(data_list):
        try:
            train_data.append((json.loads(item).get(text_key), {'entities':json.loads(item).get(offset_key)}))
        except:
            pass
        
    return train_data


def _get_tags_and_tokens_collection(path, ignore_tag_order=False, encoding="UTF-8"):
    unique_tags = set()
    tags_collection, tokens_collection = [], []
    tags_df = pd.read_csv(path/'tags.csv', encoding=encoding, dtype='str')
    tokens_df = pd.read_csv(path/'tokens.csv', encoding=encoding, dtype='str')

    for i,tags in tags_df.iterrows():
        if ignore_tag_order: tags = [x.split("-")[-1] for x in tags.dropna()]
        else: tags = [x for x in tags.dropna()]
        unique_tags.update(tags)
        tags_collection.append(tags)

    for i,tokens in tokens_df.iterrows():
        tokens_collection.append(list(tokens.dropna()))

    return tags_collection, tokens_collection, unique_tags


class _NERData:
    """
    #     Prepares a data object
    #
    #     =====================   ===========================================
    #     **Argument**            **Description**
    #     ---------------------   -------------------------------------------
    #     dataset_type            Required string. ['ner_json', 'IOB', 'BILUO']
    #     ---------------------   -------------------------------------------
    #     address_tag             Optional dict. Address field/tag name
    #                             in the training data.
    #     ---------------------   -------------------------------------------
    #     val_split_pct           Optional Float. Percentage of training data to keep
    #                             as validation. The default value is 0.1.
    #     =====================   ===========================================
    #     returns: A list [text,{entities},text,{entities}] that can be ingested by ``EntityRecognizer``.
    """

    def __init__(self, dataset_type, path, batch_size, class_mapping=None, seed=42, val_split_pct=0.1, encoding="UTF-8"):
        self.dataset_type = dataset_type
        self.path = path
        self.data = None
        self.backbone = None
        self.batch_size = batch_size
        self.class_mapping = class_mapping
        self.seed = seed
        self.val_split_pct = val_split_pct
        self.encoding = encoding
        self.prepare_data_for_spacy()

    def show_batch(self):
        return self.data.show_batch()

    def get_data_object(self):
        return self.data

    def prepare_data_for_transformer(self):
        unique_tags = set()
        path = Path(self.path)
        tags_collection, tokens_collection = [], []
        if self.class_mapping:
            address_tag = self.class_mapping.get('address_tag')
        else:
            address_tag = 'Address'

        if self.dataset_type == 'ner_json':
            # converts json schema to a list of list of tokens and tags
            # json - schema:
            # ----------
            # {"id": 1, "text": "Officers were dispatched ...", "labels": [[30, 38, "Crime"], [45, 92, "Address"], ...]}
            # ----------
            # converted form:
            # [('Officers were dispatched to a', 'O'),
            #  ('robbery', 'Crime'),
            #  ('of the', 'O'),
            #  ('Associated Bank in the 1500 block of W Broadway', 'Address')]
            with open(path, 'r', encoding=self.encoding) as f:
                data_list = f.readlines()

            data_list = [json.loads(item) for item in data_list]

            for row in data_list:
                prev_start = 0
                tmp_tags_list, tmp_tokens_list = [], []
                text, labels = row["text"], row["labels"]
                for item in sorted(labels, key=lambda x: x[0]):
                    c_text = text[prev_start: item[0]].strip()
                    tmp_tags_list.append("O")
                    unique_tags.add("O")
                    tmp_tokens_list.append(c_text)
                    c_text = text[item[0]:item[1]].strip()
                    tmp_tags_list.append(item[2])
                    tmp_tokens_list.append(c_text)
                    unique_tags.add(item[2])
                    prev_start = item[1]

                tags_collection.append(tmp_tags_list)
                tokens_collection.append(tmp_tokens_list)

        elif self.dataset_type in ['BIO', 'IOB', 'LBIOU', 'BILUO']:
            tags_collection, tokens_collection, unique_tags = _get_tags_and_tokens_collection(
                path, ignore_tag_order=True, encoding=self.encoding)
        else:
            error_message = (f"Wrong argument - {self.dataset_type} supplied for `dataset_type` parameter. "
                             "Valid values are - 'ner_json', 'BIO', 'IOB', 'LBIOU' and 'BILUO'")
            raise Exception(error_message)

        self.data = TextDataObject.prepare_data_for_entity_recognition(
            tokens_collection=tokens_collection, tags_collection=tags_collection, address_tag=address_tag,
            unique_tags=unique_tags, seed=self.seed, batch_size=self.batch_size, val_split_pct=self.val_split_pct)
        self.backbone = "transformers"

    def prepare_data_for_spacy(self):
        if not HAS_SPACY:
            _raise_spacy_import_error()

        random.seed(self.seed)
        v_list = spacy.__version__.split('.')
        version = sum([int(j)*10**(2*i) for i,j in enumerate(v_list[::-1])])
        if version < 20108: #checking spacy version
            error_message = ("Entity recognition model needs spacy version 2.1.8 or higher." 
                             f"Your current spacy version is {spacy.__version__}, please update using \'pip install'")
            return logging.error(error_message)

        path = Path(self.path)
        train_data = []

        if self.class_mapping:
            address_tag = self.class_mapping.get('address_tag')
        else:
            address_tag = 'Address'

        if self.dataset_type == 'ner_json':
            train_data = _from_json(path=path, encoding=self.encoding)
            path = path.parent
        elif self.dataset_type == 'BIO' or self.dataset_type == 'IOB':
            tags_collection, tokens_collection, _ = _get_tags_and_tokens_collection(path, encoding=self.encoding)
            train_data = _from_iob_tags(tags_collection=tags_collection, tokens_collection=tokens_collection)
        elif self.dataset_type == 'LBIOU' or self.dataset_type == 'BILUO':
            nlp=spacy.blank('en')
            tags_collection, tokens_collection, _ = _get_tags_and_tokens_collection(path, encoding=self.encoding)

            for tags, tokens in zip(tags_collection, tokens_collection):
                try:
                    tags = _iob_to_biluo(tags)

                    doc = spacy.tokens.doc.Doc(
                    nlp.vocab, words = tokens, spaces = [True]*(len(tokens)-1)+[False])
                    # run the standard pipeline against it
                    for name, proc in nlp.pipeline:
                        doc = proc(doc)
                    text = ' '.join(tokens)
                    tags = _offsets_from_biluo_tags(doc, tags)
                    train_data.append((text,{'entities':tags}))
                except:
                    pass
        else:
            error_message = (f"Wrong argument - {self.dataset_type} supplied for `dataset_type` parameter. "
                             "Valid values are - 'ner_json', 'BIO', 'IOB', 'LBIOU' and 'BILUO'")
            raise Exception(error_message)

        self.data = spaCyNERDatabunch(train_data, val_split_pct=self.val_split_pct, batch_size=self.batch_size,
                                      address_tag=address_tag, test_ds=None)
        self.data.path = path
        self.backbone = "spacy"


class _spaCyNERItemlist():
    """
    Creates a dataset to store data within ``ner_databunch`` object.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    batch_size              Batch size. 
    ---------------------   -------------------------------------------
    data                    Required: list of tuple containing text and its entities.
    =====================   ===========================================

    :returns: dataset.
    """
    
    def __init__(self, batch_size, data):
        self.batch_size = batch_size
        self.entities = list({i[2] for i in pd.concat([pd.Series(i['entities']) for i in [o[1] for o in data]])}) ##Extracting all the unique entity names from input json
        self.data = data
        self.x = [o[0] for o in data]
        self.y = [o[1] for o in data]

    def __getitem__(self, i):
        return self.data[i]

    def __len__(self):
        return len(self.data)

    def _random_batch(self, data):
        res = []
        for j in range(self.batch_size): 
            res.append(random.choice(data)) 
        return res    
    
    def _entities_to_dataframe(self, item):
        """
        This function is used to create pandas dataframe from training input data json.
        """
        text = item[0]
        df = pd.DataFrame(item[1].get('entities'))

        out_dict = {}
        if len(df):
            for x in df[2].unique(): out_dict[x] = df[df[2] == x][[0, 1]].values.tolist()

        out = {}
        out['text'] = text
        for key in out_dict.keys():
            for tpl in out_dict.get(key):
                if out.get(key) == None:
                    out[key] = []
                out[key].append(text[tpl[0]:tpl[1]])
        return pd.Series(out)

    def show_batch(self):
        """
        This function shows a batch from the _spaCyNERItemlist.
        """
        data = self._random_batch(self.data)
        lst = []
        for item in data:
            lst.append(self._entities_to_dataframe(item))
        batch_df = pd.concat(lst,axis=1,sort=True).T

        text = batch_df['text']
        batch_df.drop('text', axis=1, inplace=True)
        batch_df.insert(loc=0, column='text', value=text)
        return batch_df.fillna('')


class spaCyNERDatabunch():


    """
    Creates a databunch object.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    ds                      Required data list returned from _ner_prepare_data(). 
    ---------------------   -------------------------------------------
    val_split_pct           Optional float. Percentage of training data to keep
                            as validation.
                            The default Value is 0.1.
    ---------------------   -------------------------------------------
    batch_size              Optional integer. Batch size
                            The default value is 5.                       
    =====================   ===========================================

    :returns: dataset
    """

    def __init__(self, ds, val_split_pct, batch_size, test_ds=None,address_tag=None):
        random.shuffle(ds)
        # creating an _spaCyNERItemlist with training dataset
        self.train_ds = _spaCyNERItemlist(batch_size,data = ds[:int(len(ds)*(1-val_split_pct))])
        # creating an _spaCyNERItemlist with validation dataset
        self.val_ds = _spaCyNERItemlist(batch_size,data = ds[int(len(ds)*(1-val_split_pct)):])
        self.entities=list(set(self.train_ds.entities).union(set(self.val_ds.entities)))
        self._address_tag=address_tag
        self._has_address=True
        self.batch_size=batch_size
        if self.batch_size>len(self.train_ds):
            error_message = (f"Number of training data items ({len(self.train_ds)}) "
                             f"is less than the batch size ({self.batch_size}). "
                             "Please get more training data or lower the batch size")
            logging.error(error_message)
        if self._address_tag not in self.entities:
            self._has_address=False
            logging.warning("No Address tag found in your data.\n\
                1. If your data has an address field, pass your address field name as address tag in class mapping \n\
                e.g. - data=prepare_data(dataset_type=ds_type,path=training_data_folder,\n\t\t\t\
                    class_mapping={address_tag:address_field_name})\n\
                2. Else no action is required, if your data does not have any address information.")

    def show_batch(self):
        return self.train_ds.show_batch()



"""
urllib parsing helpers to help figure out of the URL returns a file.
"""
import os
import re
import unicodedata
from urllib.parse import urlparse, urlsplit, urljoin
from urllib.parse import urlunsplit, unquote, quote


# Check https://regex101.com/r/A326u1/5 for reference
DOMAIN_FORMAT = re.compile(
    r"(?:^(\w{1,255}):(.{1,255})@|^)" # http basic authentication [optional]
    r"(?:(?:(?=\S{0,253}(?:$|:))" # check full domain length to be less than or equal to 253 (starting after http basic auth, stopping before port)
    r"((?:[a-z0-9](?:[a-z0-9-]{0,61}[a-z0-9])?\.)+" # check for at least one subdomain (maximum length per subdomain: 63 characters), dashes in between allowed
    r"(?:[a-z0-9]{1,63})))" # check for top level domain, no dashes allowed
    r"|localhost)" # accept also "localhost" only
    r"(:\d{1,5})?", # port [optional]
    re.IGNORECASE
)
SCHEME_FORMAT = re.compile(
    r"^(http|hxxp|ftp|fxp)s?$", # scheme: http(s) or ftp(s)
    re.IGNORECASE
)

def validate_url(url: str):
    url = url.strip()

    if not url:
        raise Exception("No URL specified")

    if len(url) > 2048:
        raise Exception("URL exceeds its maximum length of 2048 characters (given length={})".format(len(url)))

    result = urlparse(url)
    scheme = result.scheme
    domain = result.netloc

    if not scheme:
        raise Exception("No URL scheme specified")

    if not re.fullmatch(SCHEME_FORMAT, scheme):
        raise Exception("URL scheme must either be http(s) or ftp(s) (given scheme={})".format(scheme))

    if not domain:
        raise Exception("No URL domain specified")

    if not re.fullmatch(DOMAIN_FORMAT, domain):
        raise Exception("URL domain malformed (domain={})".format(domain))

    return url
#--------------------------------------------------------------------------
def _normalize_url(url, charset='utf-8'):
    """ Normalizes a URL. Based on http://code.google.com/p/url-normalize."""
    def _clean(string):
        string = str(unquote(string), 'utf-8', 'replace')
        return unicodedata.normalize('NFC', string).encode('utf-8')

    default_port = {
        'ftp': 21,
        'telnet': 23,
        'http': 80,
        'gopher': 70,
        'news': 119,
        'nntp': 119,
        'prospero': 191,
        'https': 443,
        'snews': 563,
        'snntp': 563,
    }

    # if there is no scheme use http as default scheme
    if url[0] not in ['/', '-'] and ':' not in url[:7]:
        url = 'http://' + url


    # shebang urls support
    url = url.replace('#!', '?_escaped_fragment_=')

    # splitting url to useful parts
    scheme, auth, path, query, fragment = urlsplit(url.strip())
    (userinfo, host, port) = re.search('([^@]*@)?([^:]*):?(.*)', auth).groups()

    # Always provide the URI scheme in lowercase characters.
    scheme = scheme.lower()

    # Always provide the host, if any, in lowercase characters.
    host = host.lower()
    if host and host[-1] == '.':
        host = host[:-1]
    # take care about IDN domains
    host = host.decode(charset).encode('idna')  # IDN -> ACE

    # Only perform percent-encoding where it is essential.
    # Always use uppercase A-through-F characters when percent-encoding.
    # All portions of the URI must be utf-8 encoded NFC from Unicode strings
    path = quote(_clean(path), "~:/?#[]@!$&'()*+,;=")
    fragment = quote(_clean(fragment), "~")

    # note care must be taken to only encode & and = characters as values
    query = "&".join(["=".join([quote(_clean(t), "~:/?#[]@!$'()*+,;=") \
                                for t in q.split("=", 1)]) for q in query.split("&")])

    # Prevent dot-segments appearing in non-relative URI paths.
    if scheme in ["", "http", "https", "ftp", "file"]:
        output = []
        for part in path.split('/'):
            if part == "":
                if not output:
                    output.append(part)
            elif part == ".":
                pass
            elif part == "..":
                if len(output) > 1:
                    output.pop()
            else:
                output.append(part)
        if part in ["", ".", ".."]:
            output.append("")
        path = '/'.join(output)

    # For schemes that define a default authority, use an empty authority if
    # the default is desired.
    if userinfo in ["@", ":@"]:
        userinfo = ""

    # For schemes that define an empty path to be equivalent to a path of "/",
    # use "/".
    if path == "" and scheme in ["http", "https", "ftp", "file"]:
        path = "/"

    # For schemes that define a port, use an empty port if the default is
    # desired
    if port and scheme in list(default_port.keys()):
        if port.isdigit():
            port = str(int(port))
            if int(port) == default_port[scheme]:
                port = ''

    # Put it all back together again
    auth = (userinfo or "") + host
    if port:
        auth += ":" + port
    if url.endswith("#") and query == "" and fragment == "":
        path += "#"
    return urlunsplit((scheme, auth, path, query, fragment))
#--------------------------------------------------------------------------
def _parse_hostname(url, include_port=False):
    """ Parses the hostname out of a URL."""
    parsed_url = urlparse((url))
    return parsed_url.netloc if include_port else parsed_url.hostname
#--------------------------------------------------------------------------
def _is_http_url(url):
    if url:
        return urlparse(url).scheme in ['http', 'https']
    return False
#--------------------------------------------------------------------------
def _unpack(obj_or_seq, key=None, flatten=False):
    """ Turns a list of single item dicts in a list of the dict's values."""

    # The trivial case (passed in None, return None)
    if not obj_or_seq:
        return None

    # We assume it's a sequence
    new_list = []
    for obj in obj_or_seq:
        value = _unpack_obj(obj, key, flatten)
        new_list.extend(value)

    return new_list
#--------------------------------------------------------------------------
def _unpack_obj(obj, key=None, flatten=False):
    try:
        if key:
            value = [obj.get(key)]
        else:
            value = list(obj.values())
    except AttributeError:
        value = [obj]

    # Flatten any lists if directed to do so
    if value and flatten:
        value = [item for sublist in value for item in sublist]

    return value
#--------------------------------------------------------------------------
def _remove_non_ascii(s):
    return ''.join(i for i in s if ord(i) < 128)
#--------------------------------------------------------------------------
def _tostr(obj):
    if not obj:
        return ''
    if isinstance(obj, list):
        return ', '.join(map(_tostr, obj))
    return str(obj)
#--------------------------------------------------------------------------
def _filename_from_url(url):
    """:return: detected filename or None"""
    fname = os.path.basename(urlparse(url).path)
    if len(fname.strip(" \n\t.")) == 0 or \
       len(fname.strip(" \n\t.").split('.')) == 1:
        return None
    return fname
#--------------------------------------------------------------------------
def _filename_from_headers(headers):
    """Detect filename from Content-Disposition headers if present.
    http://greenbytes.de/tech/tc2231/

    :param: headers as dict, list or string
    :return: filename from content-disposition header or None
    """
    if type(headers) == str:
        headers = headers.splitlines()
    if type(headers) == list:
        headers = dict([x.split(':', 1) for x in headers])
    cdisp = headers.get("Content-Disposition")
    if not cdisp:
        return None
    cdtype = cdisp.split(';')
    if len(cdtype) == 1:
        return None
    if cdtype[0].strip().lower() not in ('inline', 'attachment'):
        return None
    # several filename params is illegal, but just in case
    fnames = [x for x in cdtype[1:] if x.strip().startswith('filename=')]
    if len(fnames) > 1:
        return None
    name = fnames[0].split('=')[1].strip(' \t"')
    name = os.path.basename(name)
    if not name:
        return None
    return name



# Code borrowed from https://github.com/justchenhao/STANet
# BSD 2-Clause License

# Copyright (c) 2020, justchenhao
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import torch.nn as nn
import torch
import torch.nn.functional as F
import torch.utils.model_zoo as model_zoo
import math
from fastai.basic_train import Learner
from fastai.torch_core import split_model_idx, flatten_model
from .._utils.segmentation_loss_functions import dice
from fastai.basic_train import LearnerCallback
import numpy as np
from torch import optim
from functools import partial
AdamW = partial(optim.Adam, betas=(0.5,0.99))


class F_mynet3(nn.Module):
    def __init__(self, backbone='resnet18',in_c=3, f_c=64, output_stride=8, **kwargs):
        self.in_c = in_c
        super(F_mynet3, self).__init__()
        self.module = mynet3(backbone=backbone, output_stride=output_stride, f_c=f_c, in_c=self.in_c, **kwargs)
    def forward(self, input):
        return self.module(input)


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)

model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
}

def ResNet34(output_stride, BatchNorm=nn.BatchNorm2d, pretrained=True, in_c=3):
    """
    output, low_level_feat:
    512, 64
    """
    # print(in_c)
    model = ResNet(BasicBlock, [3, 4, 6, 3], output_stride, BatchNorm, in_c=in_c)
    if in_c != 3:
        pretrained = False
    if pretrained:
        model._load_pretrained_model(model_urls['resnet34'])
    return model

def ResNet18(output_stride, BatchNorm=nn.BatchNorm2d, pretrained=True, in_c=3):
    """
    output, low_level_feat:
    512, 256, 128, 64, 64
    """
    model = ResNet(BasicBlock, [2, 2, 2, 2], output_stride, BatchNorm, in_c=in_c)
    if in_c !=3:
        pretrained=False
    if pretrained:
        model._load_pretrained_model(model_urls['resnet18'])
    return model

def ResNet50(output_stride, BatchNorm=nn.BatchNorm2d, pretrained=True, in_c=3):
    """
    output, low_level_feat:
    2048, 256
    """
    model = ResNet(Bottleneck, [3, 4, 6, 3], output_stride, BatchNorm, in_c=in_c)
    if in_c !=3:
        pretrained=False
    if pretrained:
        model._load_pretrained_model(model_urls['resnet50'])
    return model


def ResNet101(output_stride, BatchNorm=nn.BatchNorm2d, pretrained=True, in_c=3):
    model = ResNet(Bottleneck, [3, 4, 23, 3], output_stride, BatchNorm, in_c=in_c)
    if in_c !=3:
        pretrained=False
    if pretrained:
        model._load_pretrained_model(model_urls['resnet101'])
    return model


def ResNet152(output_stride, BatchNorm=nn.BatchNorm2d, pretrained=True, in_c=3):
    model = ResNet(Bottleneck, [3, 4, 36, 3], output_stride, BatchNorm, in_c=in_c)
    if in_c != 3:
        pretrained=False
    if pretrained:
        model._load_pretrained_model(model_urls['resnet152'])
    return model


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, dilation=1, downsample=None, BatchNorm=None):
        super(BasicBlock, self).__init__()

        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=3, stride=stride,
                               dilation=dilation, padding=dilation, bias=False)
        self.bn1 = BatchNorm(planes)

        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = BatchNorm(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, dilation=1, downsample=None, BatchNorm=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = BatchNorm(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               dilation=dilation, padding=dilation, bias=False)
        self.bn2 = BatchNorm(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = BatchNorm(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride
        self.dilation = dilation

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self,  block, layers, output_stride, BatchNorm, pretrained=True, in_c=3):

        self.inplanes = 64
        self.in_c = in_c
        # print('in_c: ',self.in_c)
        super(ResNet, self).__init__()
        blocks = [1, 2, 4]
        if output_stride == 32:
            strides = [1, 2, 2, 2]
            dilations = [1, 1, 1, 1]
        elif output_stride == 16:
            strides = [1, 2, 2, 1]
            dilations = [1, 1, 1, 2]
        elif output_stride == 8:
            strides = [1, 2, 1, 1]
            dilations = [1, 1, 2, 4]
        elif output_stride == 4:
            strides = [1, 1, 1, 1]
            dilations = [1, 2, 4, 8]
        else:
            raise NotImplementedError

        # Modules
        self.conv1 = nn.Conv2d(self.in_c, 64, kernel_size=7, stride=2, padding=3,
                                bias=False)
        self.bn1 = BatchNorm(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0], stride=strides[0], dilation=dilations[0], BatchNorm=BatchNorm)
        self.layer2 = self._make_layer(block, 128, layers[1], stride=strides[1], dilation=dilations[1], BatchNorm=BatchNorm)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=strides[2], dilation=dilations[2], BatchNorm=BatchNorm)
        self.layer4 = self._make_MG_unit(block, 512, blocks=blocks, stride=strides[3], dilation=dilations[3], BatchNorm=BatchNorm)
        # self.layer4 = self._make_layer(block, 512, layers[3], stride=strides[3], dilation=dilations[3], BatchNorm=BatchNorm)
        self._init_weight()


    def _make_layer(self, block, planes, blocks, stride=1, dilation=1, BatchNorm=None):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                BatchNorm(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, dilation, downsample, BatchNorm))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, dilation=dilation, BatchNorm=BatchNorm))

        return nn.Sequential(*layers)

    def _make_MG_unit(self, block, planes, blocks, stride=1, dilation=1, BatchNorm=None):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                BatchNorm(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, dilation=blocks[0]*dilation,
                            downsample=downsample, BatchNorm=BatchNorm))
        self.inplanes = planes * block.expansion
        for i in range(1, len(blocks)):
            layers.append(block(self.inplanes, planes, stride=1,
                                dilation=blocks[i]*dilation, BatchNorm=BatchNorm))

        return nn.Sequential(*layers)

    def forward(self, input):
        x = self.conv1(input)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)   # | 4
        x = self.layer1(x)  # | 4
        low_level_feat2 = x  # | 4
        x = self.layer2(x)  # | 8
        low_level_feat3 = x
        x = self.layer3(x)  # | 16
        low_level_feat4 = x
        x = self.layer4(x)  # | 32
        return x, low_level_feat2, low_level_feat3, low_level_feat4

    def _init_weight(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _load_pretrained_model(self, model_path):
        pretrain_dict = model_zoo.load_url(model_path)
        model_dict = {}
        state_dict = self.state_dict()
        for k, v in pretrain_dict.items():
            if k in state_dict:
                model_dict[k] = v
        state_dict.update(model_dict)
        self.load_state_dict(state_dict)



def build_backbone(backbone, output_stride, BatchNorm, in_c=3):
    if backbone == 'resnet152':
        return ResNet152(output_stride, BatchNorm, in_c=in_c)    
    if backbone == 'resnet101':
        return ResNet101(output_stride, BatchNorm, in_c=in_c)    
    if backbone == 'resnet50':
        return ResNet50(output_stride, BatchNorm, in_c=in_c)
    elif backbone == 'resnet34':
        return ResNet34(output_stride, BatchNorm, in_c=in_c)
    elif backbone == 'resnet18':
        return ResNet18(output_stride, BatchNorm, in_c=in_c)
    else:
        raise NotImplementedError


def define_F(in_c, f_c, type='unet', backbone='resnet18', **kwargs):
    if type == 'mynet3':
        # print("using mynet3 backbone")
        return F_mynet3(backbone=backbone, in_c=in_c, f_c=f_c, output_stride=32, **kwargs)
    else:
        NotImplementedError('no such F type!')

def weights_init(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        nn.init.normal_(m.weight.data, 0.0, 0.02)
    elif classname.find('BatchNorm') != -1:
        nn.init.normal_(m.weight.data, 1.0, 0.02)
        nn.init.constant_(m.bias.data, 0)



class CDSA(nn.Module):
    """self attention module for change detection

    """
    def __init__(self, in_c, ds=1, mode='BAM'):
        super(CDSA, self).__init__()
        self.in_C = in_c
        self.ds = ds
        # print('ds: ', self.ds)
        self.mode = mode
        if self.mode == 'BAM':
            self.Self_Att = BAM(self.in_C, ds=self.ds)
        elif self.mode == 'PAM':
            self.Self_Att = PAM(in_channels=self.in_C, out_channels=self.in_C, sizes=[1,2,4,8],ds=self.ds)
        self.apply(weights_init)

    def forward(self, x1, x2):
        height = x1.shape[3]
        x = torch.cat((x1, x2), 3)
        x = self.Self_Att(x)
        return x[:,:,:,0:height], x[:,:,:,height:]



class DR(nn.Module):
    def __init__(self, in_d, out_d):
        super(DR, self).__init__()
        self.in_d = in_d
        self.out_d = out_d
        self.conv1 = nn.Conv2d(self.in_d, self.out_d, 1, bias=False)
        self.bn1 = nn.BatchNorm2d(self.out_d)
        self.relu = nn.ReLU()

    def forward(self, input):
        x = self.conv1(input)
        x = self.bn1(x)
        x = self.relu(x)
        return x


class Decoder(nn.Module):
    def __init__(self, fc, backbone, BatchNorm):
        super(Decoder, self).__init__()
        self.fc = fc
        if backbone in ['resnet18', 'resnet34']:
            self.dr2 = DR(64, 96)
            self.dr3 = DR(128, 96)
            self.dr4 = DR(256, 96)
            self.dr5 = DR(512, 96)
        else:
            self.dr2 = DR(256, 96)
            self.dr3 = DR(512, 96)
            self.dr4 = DR(1024, 96)
            self.dr5 = DR(2048, 96)
        self.last_conv = nn.Sequential(nn.Conv2d(384, 256, kernel_size=3, stride=1, padding=1, bias=False),
                                       BatchNorm(256),
                                       nn.ReLU(),
                                       nn.Dropout(0.5),
                                       nn.Conv2d(256, self.fc, kernel_size=1, stride=1, padding=0, bias=False),
                                       BatchNorm(self.fc),
                                       nn.ReLU(),
                                       )

        self._init_weight()


    def forward(self, x,low_level_feat2, low_level_feat3, low_level_feat4):

        # x1 = self.dr1(low_level_feat1)
        x2 = self.dr2(low_level_feat2)
        x3 = self.dr3(low_level_feat3)
        x4 = self.dr4(low_level_feat4)
        x = self.dr5(x)
        x = F.interpolate(x, size=x2.size()[2:], mode='bilinear', align_corners=True)
        # x2 = F.interpolate(x2, size=x3.size()[2:], mode='bilinear', align_corners=True)
        x3 = F.interpolate(x3, size=x2.size()[2:], mode='bilinear', align_corners=True)
        x4 = F.interpolate(x4, size=x2.size()[2:], mode='bilinear', align_corners=True)

        x = torch.cat((x, x2, x3, x4), dim=1)

        x = self.last_conv(x)

        return x

    def _init_weight(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                torch.nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()


def build_decoder(fc, backbone, BatchNorm):
    return Decoder(fc, backbone, BatchNorm)


from ._arcgis_model import _change_tail

class mynet3(nn.Module):
    def __init__(self, backbone='resnet18', output_stride=16, f_c=64, freeze_bn=False, in_c=3, data=None):
        super(mynet3, self).__init__()
        # print('arch: mynet3')
        BatchNorm = nn.BatchNorm2d
        self.backbone = build_backbone(backbone, output_stride, BatchNorm, in_c)
        if data is not None:
            if data._is_multispectral:
                self.backbone = _change_tail(self.backbone, data)
        self.decoder = build_decoder(f_c, backbone, BatchNorm)

        if freeze_bn:
            self.freeze_bn()

    def forward(self, input):
        x, f2, f3, f4 = self.backbone(input)
        x = self.decoder(x, f2, f3, f4)
        return x

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()




class _PAMBlock(nn.Module):
    '''
    The basic implementation for self-attention block/non-local block
    Input/Output:
        N * C  *  H  *  (2*W)
    Parameters:
        in_channels       : the dimension of the input feature map
        key_channels      : the dimension after the key/query transform
        value_channels    : the dimension after the value transform
        scale             : choose the scale to partition the input feature maps
        ds                : downsampling scale
    '''
    def __init__(self, in_channels, key_channels, value_channels, scale=1, ds=1):
        super(_PAMBlock, self).__init__()
        self.scale = scale
        self.ds = ds
        self.pool = nn.AvgPool2d(self.ds)
        self.in_channels = in_channels
        self.key_channels = key_channels
        self.value_channels = value_channels

        self.f_key = nn.Sequential(
            nn.Conv2d(in_channels=self.in_channels, out_channels=self.key_channels,
                kernel_size=1, stride=1, padding=0),
            nn.BatchNorm2d(self.key_channels)
        )
        self.f_query = nn.Sequential(
            nn.Conv2d(in_channels=self.in_channels, out_channels=self.key_channels,
                kernel_size=1, stride=1, padding=0),
            nn.BatchNorm2d(self.key_channels)
        )
        self.f_value = nn.Conv2d(in_channels=self.in_channels, out_channels=self.value_channels,
            kernel_size=1, stride=1, padding=0)



    def forward(self, input):
        x = input
        if self.ds != 1:
            x = self.pool(input)
        # input shape: b,c,h,2w
        batch_size, c, h, w = x.size(0), x.size(1), x.size(2), x.size(3)//2


        local_y = []
        local_x = []
        step_h, step_w = h//self.scale, w//self.scale
        for i in range(0, self.scale):
            for j in range(0, self.scale):
                start_x, start_y = i*step_h, j*step_w
                end_x, end_y = min(start_x+step_h, h), min(start_y+step_w, w)
                if i == (self.scale-1):
                    end_x = h - (h % step_h)
                if j == (self.scale-1):
                    end_y = w - (w % step_h)
                local_x += [start_x, end_x]
                local_y += [start_y, end_y]

        value = self.f_value(x)
        query = self.f_query(x)
        key = self.f_key(x)

        value = torch.stack([value[:, :, :, :w], value[:,:,:,w:]], 4)  # B*N*H*W*2
        query = torch.stack([query[:, :, :, :w], query[:,:,:,w:]], 4)  # B*N*H*W*2
        key = torch.stack([key[:, :, :, :w], key[:,:,:,w:]], 4)  # B*N*H*W*2


        local_block_cnt = 2*self.scale*self.scale

        #  self-attention func
        def func(value_local, query_local, key_local):
            batch_size_new = value_local.size(0)
            h_local, w_local = value_local.size(2), value_local.size(3)
            value_local = value_local.contiguous().view(batch_size_new, self.value_channels, -1)

            query_local = query_local.contiguous().view(batch_size_new, self.key_channels, -1)
            query_local = query_local.permute(0, 2, 1)
            key_local = key_local.contiguous().view(batch_size_new, self.key_channels, -1)

            sim_map = torch.bmm(query_local, key_local)  # batch matrix multiplication
            sim_map = (self.key_channels**-.5) * sim_map
            sim_map = F.softmax(sim_map, dim=-1)

            context_local = torch.bmm(value_local, sim_map.permute(0,2,1))
            # context_local = context_local.permute(0, 2, 1).contiguous()
            context_local = context_local.view(batch_size_new, self.value_channels, h_local, w_local, 2)
            return context_local

        #  Parallel Computing to speed up
        #  reshape value_local, q, k
        v_list = [value[:,:,local_x[i]:local_x[i+1],local_y[i]:local_y[i+1]] for i in range(0, local_block_cnt, 2)]
        v_locals = torch.cat(v_list,dim=0)
        q_list = [query[:,:,local_x[i]:local_x[i+1],local_y[i]:local_y[i+1]] for i in range(0, local_block_cnt, 2)]
        q_locals = torch.cat(q_list,dim=0)
        k_list = [key[:,:,local_x[i]:local_x[i+1],local_y[i]:local_y[i+1]] for i in range(0, local_block_cnt, 2)]
        k_locals = torch.cat(k_list,dim=0)
        # print(v_locals.shape)
        context_locals = func(v_locals,q_locals,k_locals)

        context_list = []
        for i in range(0, self.scale):
            row_tmp = []
            for j in range(0, self.scale):
                left = batch_size*(j+i*self.scale)
                right = batch_size*(j+i*self.scale) + batch_size
                tmp = context_locals[left:right]
                row_tmp.append(tmp)
            context_list.append(torch.cat(row_tmp, 3))

        context = torch.cat(context_list, 2)
        context = torch.cat([context[:,:,:,:,0],context[:,:,:,:,1]],3)


        if self.ds !=1:
            context = F.interpolate(context, [h*self.ds, 2*w*self.ds])

        return context


class PAMBlock(_PAMBlock):
    def __init__(self, in_channels, key_channels=None, value_channels=None, scale=1, ds=1):
        if key_channels == None:
            key_channels = in_channels//8
        if value_channels == None:
            value_channels = in_channels
        super(PAMBlock, self).__init__(in_channels,key_channels,value_channels,scale,ds)


class PAM(nn.Module):
    """
        PAM module
    """

    def __init__(self, in_channels, out_channels, sizes=([1]), ds=1):
        super(PAM, self).__init__()
        self.group = len(sizes)
        self.stages = []
        self.ds = ds  # output stride
        self.value_channels = out_channels
        self.key_channels = out_channels // 8


        self.stages = nn.ModuleList(
            [self._make_stage(in_channels, self.key_channels, self.value_channels, size, self.ds)
             for size in sizes])
        self.conv_bn = nn.Sequential(
            nn.Conv2d(in_channels * self.group, out_channels, kernel_size=1, padding=0,bias=False),
            # nn.BatchNorm2d(out_channels),
        )

    def _make_stage(self, in_channels, key_channels, value_channels, size, ds):
        return PAMBlock(in_channels,key_channels,value_channels,size,ds)

    def forward(self, feats):
        priors = [stage(feats) for stage in self.stages]

        #  concat
        context = []
        for i in range(0, len(priors)):
            context += [F.interpolate(priors[i], size=feats.shape[-2:], mode='bilinear', align_corners=True)]
            
        output = self.conv_bn(torch.cat(context, 1))

        return output


class BAM(nn.Module):
    """ Basic self-attention module
    """

    def __init__(self, in_dim, ds=8, activation=nn.ReLU):
        super(BAM, self).__init__()
        self.chanel_in = in_dim
        self.key_channel = self.chanel_in //8
        self.activation = activation
        self.ds = ds  #
        self.pool = nn.AvgPool2d(self.ds)
        # print('ds: ',ds)
        self.query_conv = nn.Conv2d(in_channels=in_dim, out_channels=in_dim // 8, kernel_size=1)
        self.key_conv = nn.Conv2d(in_channels=in_dim, out_channels=in_dim // 8, kernel_size=1)
        self.value_conv = nn.Conv2d(in_channels=in_dim, out_channels=in_dim, kernel_size=1)
        self.gamma = nn.Parameter(torch.zeros(1))

        self.softmax = nn.Softmax(dim=-1)  #

    def forward(self, input):
        """
            inputs :
                x : input feature maps( B X C X W X H)
            returns :
                out : self attention value + input feature
                attention: B X N X N (N is Width*Height)
        """
        x = self.pool(input)
        m_batchsize, C, width, height = x.size()
        proj_query = self.query_conv(x).view(m_batchsize, -1, width * height).permute(0, 2, 1)  # B X C X (N)/(ds*ds)
        proj_key = self.key_conv(x).view(m_batchsize, -1, width * height)  # B X C x (*W*H)/(ds*ds)
        energy = torch.bmm(proj_query, proj_key)  # transpose check
        energy = (self.key_channel**-.5) * energy

        attention = self.softmax(energy)  # BX (N) X (N)/(ds*ds)/(ds*ds)

        proj_value = self.value_conv(x).view(m_batchsize, -1, width * height)  # B X C X N

        out = torch.bmm(proj_value, attention.permute(0, 2, 1))
        out = out.view(m_batchsize, C, width, height)

        out = F.interpolate(out, [width*self.ds,height*self.ds])
        out = out + input

        return out


class BCL(nn.Module):
    """
    batch-balanced contrastive loss
    no-change，1
    change，-1
    """

    def __init__(self, margin=2.0):
        super(BCL, self).__init__()
        self.margin = margin

    def forward(self, distance, label):
        # Not getting used.
        label[label==255] = 1
        mask = (label != 255).float()
        distance = distance * mask
        pos_num = torch.sum((label==1).float())+0.0001
        neg_num = torch.sum((label==-1).float())+0.0001

        loss_1 = torch.sum((1+label) / 2 * torch.pow(distance, 2)) /pos_num
        loss_2 = torch.sum((1-label) / 2 *
            torch.pow(torch.clamp(self.margin - distance, min=0.0), 2)
        ) / neg_num
        loss = loss_1 + loss_2
        return loss

class CD_Model(nn.Module):
    def __init__(self, netF, netA):
        super().__init__()
        self.netF = netF
        self.netA = netA
        
    def forward(self, inp1, inp2):
        feat_A = self.netF(inp1)
        feat_B = self.netF(inp2)
        # print(feat_A.shape, feat_B.shape)
        feat_A, feat_B = self.netA(feat_A, feat_B)
        
        
        dist = F.pairwise_distance(feat_A, feat_B, keepdim=True)

        dist = F.interpolate(dist, size=inp1.shape[2:], mode='bilinear', align_corners=True)

        return dist


def split_layer_groups(learn):
    layers = flatten_model(learn.model)
    for idx, l in enumerate(layers):
        if 'decoder' in layers:
            break
    learn.layer_groups = split_model_idx(learn.model, [idx])              

class METRICS():
    "Datastore"
    precision = None
    recall = None
    f1 = None
    pass

from .._utils.classified_tiles import post_process_CD, calculate_metrics


def compute_datastore(preds, targets):
    ## hardcoding 2 for binary CD
    preds, targets = post_process_CD(preds, targets, 2)
    precision, recall, f1 = calculate_metrics(preds, 
                                              targets,
                                              'per_class')
    # print(precision.shape, recall.shape, f1.shape)
    METRICS.precision = precision.mean(0)[1]
    METRICS.recall = recall.mean(0)[1]
    METRICS.f1 = f1.mean(0)[1]


def precision(preds, targets):
    if METRICS.precision is None:
        compute_datastore(preds, targets)
    return METRICS.precision


def recall(preds, targets):
    if METRICS.recall is None:
        compute_datastore(preds, targets)
    return METRICS.recall


def f1(preds, targets):
    if METRICS.f1 is None:
        compute_datastore(preds, targets)
    return METRICS.f1


def get_scores(confusion_matrix):
    """
    Returns score about:
        - overall accuracy
        - mean accuracy
        - mean IU
        - fwavacc
    For reference, please see: https://en.wikipedia.org/wiki/Confusion_matrix
    :return:
    """
    hist = confusion_matrix
    num_classes = len(confusion_matrix)
    tp = np.diag(hist)
    sum_a1 = hist.sum(axis=1)
    sum_a0 = hist.sum(axis=0)

    # ---------------------------------------------------------------------- #
    # 1. Accuracy & Class Accuracy
    # ---------------------------------------------------------------------- #
    acc = tp.sum() / (hist.sum() + np.finfo(np.float32).eps)

    # recall
    acc_cls_ = tp / (sum_a1 + np.finfo(np.float32).eps)

    # precision
    precision = tp / (sum_a0 + np.finfo(np.float32).eps)
    # ---------------------------------------------------------------------- #
    # 2. Mean IoU
    # ---------------------------------------------------------------------- #
    iu = tp / (sum_a1 + hist.sum(axis=0) - tp + np.finfo(np.float32).eps)
    mean_iu = np.nanmean(iu)

    cls_iu = dict(zip(range(num_classes), iu))

    # F1 score
    F1 = 2 * acc_cls_ * precision / (acc_cls_ + precision + np.finfo(np.float32).eps)

    scores = {'Overall_Acc': acc,
            'Mean_IoU': mean_iu}
    scores.update(cls_iu)
    scores.update({'precision_1': precision[1],
                    'recall_1': acc_cls_[1],
                    'F1_1': F1[1]})
    return scores

def __fast_hist(label_gt, label_pred, num_classes):
    """
    Collect values for Confusion Matrix
    For reference, please see: https://en.wikipedia.org/wiki/Confusion_matrix
    :param label_gt: <np.array> ground-truth
    :param label_pred: <np.array> prediction
    :return: <np.ndarray> values for confusion matrix
    """
    mask = (label_gt >= 0) & (label_gt < num_classes)
    hist = np.bincount(num_classes * label_gt[mask].astype(int) + label_pred[mask],
                        minlength=num_classes**2).reshape(num_classes, num_classes)
    return hist
# def precision_recall_score()

def calculate_author_metrics(self):
    n_classes = len(self._data.classes)
    dl = self._data.valid_dl
    #
    confusion_matrix = np.zeros((n_classes, n_classes))
    for x, y in dl:
        with torch.no_grad():
            predictions = self.learn.model(*x).detach().to('cpu')
        predictions, y = post_process_CD(predictions, 
                                         y, 
                                         n_classes=n_classes)

        # print(y.shape, predictions.shape)
        predictions = predictions.argmax(1).numpy().flatten()
        y = y.cpu().numpy().flatten()
        # print(y.shape, predictions.shape)
        confusion_matrix += __fast_hist(y, predictions, n_classes)

    return get_scores(confusion_matrix)


def get_learner(data, backbone, self_attention_type='PAM'):
    netF = define_F(in_c=3, 
                    f_c=64, 
                    type='mynet3', 
                    backbone=backbone, 
                    data=data if data._is_multispectral else None)
    netA = CDSA(in_c=64, ds=1, mode=self_attention_type)
    lossf = BCL()
    learn = Learner(
        data=data, 
        model=CD_Model(netF, netA), 
        loss_func=lossf,
        metrics=[precision, recall, f1],
        opt_func=AdamW
        )
    learn.callbacks.append(ResetMetricCallback(learn))
    split_layer_groups(learn)
    learn.freeze()
    return learn


class ResetMetricCallback(LearnerCallback):
    def on_batch_end(self, **kwargs):
        METRICS.precision = None
        METRICS.recall = None
        METRICS.f1 = None
    


import os
import time
import sys

from arcgis.widgets._mapview._raster import *
from arcgis.widgets._mapview._raster._numpy_utils import *
from arcgis.widgets._mapview._raster._jupyter_utils import *

class RasterOverlay:
    def __init__(self,
                 id_: str,
                 img_url: str,
                 extent: dict,
                 opacity: float):
        self.id = id_
        self.img_url = img_url
        self.extent = extent
        self.opacity = opacity

    def as_dict(self) -> dict:
        return { href: self.img_url,
                 extent: self.extent,
                 opacity: self.opacity }

class RasterData:
    def __init__(self, 
                 img_data,
                 extent: dict,
                 id_: str):
        self.img_data = img_data
        self.extent = extent
        self.id = id_

class LocalRasterOverlayManager:
    """
    .. warning::
        Overlying local rasters on a ``MapView`` instance have the following 
        limitations:
        
        - Local raster overlays do not persist beyond the notebook session on
          published web maps/web scenes -- you would need to seperately publish
          these local rasters.
        
        - The entire raster image data is placed on the MapView's canvas with 
          no performance optimizations. This means no pyramids, no dynamic 
          downsampling, etc. Please be mindful of the size of the local raster 
          and your computer's hardware limitations.
        
        - Pixel values and projections are not guaranteed to be accurate, 
          especially when the local raster's Spatial Reference doesn't 
          reproject accurately to Web Mercator (what the ``MapView`` 
          widget uses).
    """

    def __init__(self, mapview):
        self._mapview = mapview
        self._attempt_infer_file_format()

    _overlays = []

    _file_format = "jpg"
    @property
    def file_format(self) -> str:
       return self._file_format

    @file_format.setter
    def file_format(self, value):
        self._file_format = value

    def overlay(self, raster):
        self._clear_old_image_overlays()
        raster_data = self.try_get_raster_data(raster)
        img_path = self._save_img_to_jupyter_accessible_dir(
            raster_data.img_data, raster.cmap,
            vmin = raster.vmin, vmax = raster.vmax,
            filename = str(raster_data.id) + "." + self.file_format)

        extent = self._make_extent_valid(raster.extent)

        img_url = self._get_jupyter_accessible_url(img_path)

        raster_overlay = RasterOverlay(raster_data.id, img_url, 
                                       extent, raster.opacity)
        self._mapview._add_overlay(raster_overlay)
        self._overlays.append({"id": raster_data.id,
                               "image": raster,
                               "img_path": img_path,
                               "img_url": img_url,
                               "extent": extent})

    _SUPPORTED_IMAGE_FORMATS = [".png", ".jpg", ".tif"]
    def _is_supported_image_file(self, image):
        if not isinstance(image, str):
            return False
        _, ext = os.path.splitext(image)
        return os.path.exists(image) and ext in self._SUPPORTED_IMAGE_FORMATS

    def _is_arcgis_raster(self, image):
        try:
            from arcgis.raster import Raster
            return isinstance(image, Raster)
        except ImportError:
            return False

    def try_get_raster_data(self, image) -> RasterData:
        img_data = image.read().squeeze()
        return RasterData(img_data = img_data, extent = image.extent,
                          id_ = get_hash_numpy_array(img_data))

    def remove(self, raster=None):
        image = raster
        any_removed = False
        for i in reversed(range(0, len(self._overlays))):
            overlay = self._overlays[i]
            if image is None or self._are_equal(image, overlay["image"]):
                del self._overlays[i]
                self._mapview._remove_overlay(overlay["id"])
                any_removed = True
        if any_removed:
            return True
        else:
            raise Exception(f"Could not find {image} to remove")

    def _are_equal(self, image1, image2):
        if is_numpy_array(image1):
            return get_hash_numpy_array(image1) == \
                   get_hash_numpy_array(image2)
        else:
            return image1 == image2


    def list(self):
        """
        Returns a list of rasters overlayed. Will also visually display these
        rasters in a Jupyter Notebook environment.
        """
        from IPython.display import display
        scroll_list = JupyterHorizScrollImageList()
        for overlay in self._overlays:
            scroll_list.append(src=overlay["img_url"], title=str(overlay["extent"]))
        display(scroll_list)

        return list(overlay["image"] for overlay in self._overlays)

    def _make_extent_valid(self, extent):
        if not extent:
            raise RuntimeError(f"Could not infer extent of raster! Please manually "
                               f"specify an extent")
        return extent

    def _attempt_infer_file_format(self):
        try:
            from PIL import Image
            self.file_format = "jpg"
        except ImportError:
            self.file_format = "png"

    _jupyter_notebook_dir_override = ""
    def set_current_executing_nb_dir(self, path: str):
        """To display in a notebook, rasters must be placed in the 
        ``_image_overlays`` folder in the same directory as the current
        executing notebook. This value is normally inferred. To override,
        call this function with the path to the current executing notebook.
        """
        self._jupyter_notebook_dir_override = path

    _DEFAULT_CMAP = "Greys_r" # Mirrors the style of default ArcGIS Pro

    def _save_img_to_jupyter_accessible_dir(self, img_data, cmap, vmin, vmax, filename):
        import matplotlib.pyplot as plt
        img_path = os.path.join(self._get_image_overlays_dir(), filename)
        num_bands = self._get_num_bands(img_data)
        kwargs = {}
        if vmin:
            kwargs["vmin"] = vmin
        if vmax:
            kwargs["vmax"] = vmax
        if not cmap:
            if num_bands != 1 and num_bands != 3 and num_bands != 4:
                raise Exception(f"This raster has {num_bands} bands -- Number "\
                    f"of bands must be 1 (greyscale), 3 (RGB), or 4 (RGBA). "\
                    f"Consider choosing a subset of bands, or make sure the "\
                    f"shape of the data is valid. (Shape: {img_data.shape}).")
            if num_bands == 1:
                kwargs["cmap"] = self._DEFAULT_CMAP
        else:
            if num_bands != 1:
                raise Exception(f"To use a cmap, the input raster "\
                                f"must have only 1 band, not {num_bands}")
            else:
                kwargs["cmap"] = cmap

        plt.imsave(img_path, img_data, **kwargs)

        return img_path

    def _get_num_bands(self, img_data):
        num_bands = -1
        if len(img_data.shape) == 2:
            num_bands = 1
        elif len(img_data.shape) == 3:
            num_bands = img_data.shape[2]
        return num_bands

    _image_overlays_dir_name = "_image_overlays"

    def _get_image_overlays_dir(self):
        if self._jupyter_notebook_dir_override:
            curr_jupyter_notebook_dir = self._jupyter_notebook_dir_override
        else:
            from arcgis.widgets._mapview._raster._jupyter_utils \
                import get_dir_of_curr_exec_notebook
            curr_jupyter_notebook_dir = get_dir_of_curr_exec_notebook()

        image_overlays_dir = os.path.join(curr_jupyter_notebook_dir,
                                          self._image_overlays_dir_name)
        if not os.path.isdir(image_overlays_dir):
            os.mkdir(image_overlays_dir)
        return image_overlays_dir

    def _clear_old_image_overlays(self):
        """Clears all files older than 2 days in `_image_overlays` dir"""
        dir_ = self._get_image_overlays_dir()
        now = time.time()
        for file_name in os.listdir(dir_):
            file_path = os.path.join(dir_, file_name)
            if os.stat(file_path).st_mtime < now - 2 * 86400:
                #If older than 2 days
                try:
                    os.remove(file_path)
                except Exception:
                    pass

    def _get_jupyter_accessible_url(self, image_path):
        filename = os.path.basename(image_path)
        image_url = os.path.join(self._image_overlays_dir_name, filename)
        return image_url



import os
import sys
import json
import datetime
from arcgis.gis import GIS, User, Item
from arcgis._impl.common._isd import InsensitiveDict
from arcgis._impl.common._utils import local_time_to_online
###########################################################################
class BaseTask(object):
    """
    Base Schedule Class
    """
    _url = None
    _gis = None
    _con = None
    _properties = None
    def __init__(self, url:str, gis:GIS):
        self._url = url
        self._gis = gis
    #----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}>"
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    @property
    def properties(self):
        if self._properties is None:
            params = {'f' : 'json'}
            res = self._gis._con.get(self._url, params)
            self._properties = InsensitiveDict(res)
        return self._properties
###########################################################################
class Run(BaseTask):
    """
    Represents a single run of a scheduled task.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string. The URL to the REST endpoint.
    ------------------     --------------------------------------------------------------------
    gis                    Required GIS. The GIS object.
    ==================     ====================================================================
    """
    _gis = None
    _url = None
    #----------------------------------------------------------------------
    def __init__(self, url:str, gis:GIS):
        super(Run, self)
        self._url = url
        self._gis = gis
    #----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__} @ {self.properties.runId}>"
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    def delete(self) -> bool:
        """
        Removes the Task from the System.

        :returns: Boolean

        """
        url = f"{self._url}/delete"
        params = {'f' : 'json'}
        res = self._gis._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    def update(self,
               status:str=None,
               description:str=None):
        """
        Updates the Run's Status Message and Result Message.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        status                 Optional String. The status of the run.  The allowed values are:
                               `scheduled`, `executing`, `succeeded`, `failed`, or `skipped`.
        ------------------     --------------------------------------------------------------------
        description            Optional String. Updates the descriptive message associated with the 
                               current `Run`. 
        ==================     ====================================================================

        :returns: Bool

        """
        params = {'f' : 'json'}
        status_values = ["scheduled", "executing","succeeded", "failed", "skipped"]
        if status is None and description is None:
            return False
        if status and status.lower() in status_values:
            params['status'] = status.lower()
        elif status and status.lower() not in status_values:
            raise ValueError("Invalid status")
        elif status is None:
            params['status'] = self.properties.status
        if description:
            params['result'] = description
        elif description is None:
            params['result'] = self.properties.result

        url = f"{self._url}/update"
        res = self._gis._con.post(url, params)
        if 'success' in res:
            self._properties = None
            return res['success']
        return res
###########################################################################
class Task(BaseTask):
    """
    Represents a schduled task that can be modified for a user.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string. The URL to the REST endpoint.
    ------------------     --------------------------------------------------------------------
    gis                    Required GIS. The GIS object.
    ==================     ====================================================================

    """
    _url = None
    _gis = None
    def __init__(self, url:str, gis:GIS):
        super(Task, self)
        self._url = url
        self._gis = gis
    #----------------------------------------------------------------------
    def __str__(self):
        return f"<Task @ {self.properties.id}>"
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    def delete(self) -> bool:
        """
        Removes the Task from the System.

        :returns: Boolean

        """
        url = f"{self._url}/delete"
        params = {'f' : 'json'}
        res = self._gis._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    def enable(self, enabled:bool) -> bool:
        """
        The `enable` method allows administrators to enable or disable the scheduled task..
        
        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        enabled                Required Boolean.  If True, the status of the task is set to active. 
                               If False, the task is set active to False.
        ==================     ====================================================================
        
        :returns: Bool
        
        """
        params = {'f' : 'json'}
        if enabled == True:
            url = f"{self._url}/enable"
        elif enabled == False:
            url = f"{self._url}/disable"
        else:
            raise ValueError("`enabled` must be a boolean value")
        res = self._gis._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    def start(self) -> bool:
        """
        Starts a task if it is actively running.

        :returns: Bool

        """
        return self.update(is_active=True)
    #----------------------------------------------------------------------
    def stop(self) -> bool:
        """
        Stops a task if it is actively running.

        :returns: Bool

        """
        return self.update(is_active=False)
    #----------------------------------------------------------------------
    def update(self,
               item:Item=None,
               cron:str=None,
               task_type:str=None,
               occurences:int=10,
               start_date:datetime.datetime=None,
               end_date:datetime.datetime=None,
               title:str=None,
               parameters:dict=None,
               task_url:str=None,
               is_active:bool=None) -> bool:
        """
        Updates the current Task

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Optional Item. The item to update the schedule for.
        ------------------     --------------------------------------------------------------------
        cron                   Optional String. The executution time syntax. 
        ------------------     --------------------------------------------------------------------
        task_type              Optional String. The type of task. Two valid options are `ExecuteNotebook` or `UpdateInsightsWorkbook`.
        ------------------     --------------------------------------------------------------------
        occurences             Optional Integer. The maximum number of occurrences this task should execute.
        ------------------     --------------------------------------------------------------------
        start_date             Optional Datetime. The date/time when the task will begin executing.
        ------------------     --------------------------------------------------------------------
        end_date               Optional Datetime.  The date/time when the task will stop executing.
        ------------------     --------------------------------------------------------------------
        title                  Optional String. The name of the task.
        ------------------     --------------------------------------------------------------------
        parameters             Optional Dict.  Additional key/value pairs for execution of notebooks.
        ------------------     --------------------------------------------------------------------
        task_url               Optional String. A response URL with a set of results.
        ------------------     --------------------------------------------------------------------
        is_active              Optional Bool. Determines if the tasks is currently running.
        ==================     ====================================================================

        :returns: bool or Dict on error.
        
        """
        SPECIALS = {"reboot":   '@reboot',
                    "hourly":   '0 * * * *',
                    "daily":    '0 0 * * *',
                    "weekly":   '0 0 * * 0',
                    "monthly":  '0 0 1 * *',
                    "yearly":   '0 0 1 1 *',
                    "annually": '0 0 1 1 *',
                    "midnight": '0 0 * * *'}
        params = {
            "title" : title or self.properties.title,
            "type": task_type or self.properties.type,
            "taskUrl": task_url or "",
            "parameters": parameters,
            "itemId": None,
            "minute": None,
            "hour": None,
            "dayOfMonth": None,
            "month" : None,
            "dayOfWeek": None,
            "maxOccurrences": occurences or self.properties.maxOccurrences,
            "isActive" : None,
            "f": "json"
        }
        if is_active is None:
            params.pop('isActive', None)
        else:
            params['isActive'] = json.dumps(is_active)
        if task_url is None:
            params.pop('taskUrl', None)
        if cron is None:
            params['minute'] = self.properties.cronSchedule.minute
            params['hour'] = self.properties.cronSchedule.hour
            params['dayOfMonth'] = self.properties.cronSchedule.dayOfMonth
            params['month'] = self.properties.cronSchedule.month
            params['dayOfWeek'] = self.properties.cronSchedule.dayOfWeek
        elif isinstance(cron, str) and cron in SPECIALS:
            cron = SPECIALS[cron].split(" ")
            params['minute'] = cron[0]
            params['hour'] = cron[1]
            params['dayOfMonth'] = cron[2]
            params['month'] = cron[3]
            params['dayOfWeek'] = cron[4]
        else:
            cron = cron.split(" ")
            params['minute'] = cron[0]
            params['hour'] = cron[1]
            params['dayOfMonth'] = cron[2]
            params['month'] = cron[3]
            params['dayOfWeek'] = cron[4]

        if isinstance(item, Item):
            params['itemId'] = item.itemid
        elif item:
            params['itemId'] = item
        else:
            params['itemId'] = self.properties.itemId

        if start_date:
            params['startDate'] = local_time_to_online(dt=start_date)
        else:
            params.pop('startDate', None)
        if end_date:
            params['endDate']= local_time_to_online(dt=end_date)
        else:
            params.pop('endDate', None)
        if parameters:
            params['parameters'] = json.dumps(parameters)
        elif 'parameters' in self.properties:
            params['parameters'] = self.properties.parameters
        url = f"{self._url}/update"
        res = self._gis._con.post(url, params)
        if 'success' in res:
            self._properties = None
            return res['success']
        return res
    #----------------------------------------------------------------------
    @property
    def runs(self):
        """
        Returns the Runs for the Task.  The maximum number of runs returned is 30

        :returns: List
        """
        runs = []
        url = f"{self._url}/runs"
        params = {'f' : 'json'}
        res = self._gis._con.get(url, params)
        for t in res['runs']:
            run_url = f"{self._url}/runs/{t['runId']}"
            runs.append(Run(url=run_url, gis=self._gis))
            del t
        return runs
###########################################################################
class TaskManager(object):
    """

    Provides the functions to create, update and delete scheduled tasks.

    This operation is for Enterprise configuration 10.8.1+.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string. The URL to the REST endpoint.
    ------------------     --------------------------------------------------------------------
    user                   Required User. The user to perform the Task management on.
    ------------------     --------------------------------------------------------------------
    gis                    Required GIS. The GIS object.
    ==================     ====================================================================


    """
    _tasks = None

    def __init__(self, url:str, user:User, gis:GIS):
        """initializer"""
        self._url = url
        self._user = user
        self._gis = gis
    #----------------------------------------------------------------------
    def __str__(self):
        return f"<User {self._user.username} Tasks>"
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    def search(self, item:Item=None, active:bool=None, types:str=None):
        """
        This property allows users to search for tasks based on criteria.

        ================  ===============================================================================
        **Argument**      **Description**
        ----------------  -------------------------------------------------------------------------------
        item              Optional Item. The item to query tasks about.
        ----------------  -------------------------------------------------------------------------------
        active            Optional Bool. Queries tasks based on active status.
        ----------------  -------------------------------------------------------------------------------
        types             Optional String. The type of notebook execution for the item.  This can be
                          `ExecuteNotebook`, or `UpdateInsightsWorkbook`.
        ================  ===============================================================================

        :returns: List of Tasks
        
        """        
        if item is None and \
           active is None and \
           types is None:
            return self.all
        else:
            _tasks = []
            url = f"{self._gis._portal.resturl}community/users/{self._user.username}/tasks"
            params = {
                'num' : 100,
                'start' : 1,
            }
            if item:
                params['itemId'] = item.itemid
            if not active is None:
                params['active'] = active
            if types:
                params['types'] = types
            res = self._gis._con.get(url, params)
            for t in res['tasks']:
                url = f"{self._url}/{t['id']}"
                _tasks.append(Task(url=url, gis=self._gis))
            while res['nextStart'] != -1:
                params['start'] = res['nextStart']
                res = self._gis._con.get(url, params)
                for t in res['tasks']:
                    url = f"{self._url}/{t['id']}"
                    _tasks.append(Task(url=url, gis=self._gis))
            return _tasks
        return []
    #----------------------------------------------------------------------
    def create(self,
               item:Item,
               cron:str,
               task_type:str,
               occurences:int=10,
               start_date:datetime.datetime=None,
               end_date:datetime.datetime=None,
               title:str=None,
               parameters:dict=None) -> Task:
        """
        Creates a new scheduled task for a notebook `Item`.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required Item. The item to schedule a task on.
        ------------------     --------------------------------------------------------------------
        cron                   Required String. The CRON statement. This should be in the from of:

                               `<minute> <hour> <day of month> <month> <day of week>`

                               Example to run a task weekly, use: `0 0 * * 0`
        ------------------     --------------------------------------------------------------------
        task_type              Required String. The platform to execute the notebook on.  For
                               notebook server tasks use: `ExecuteNotebook`, for Insights notebook
                               use: `UpdateInsightsWorkbook`.
        ------------------     --------------------------------------------------------------------
        occurences             Optional Integer. The total number of instance that can run at a single time.
        ------------------     --------------------------------------------------------------------
        start_date             Optional Datetime. The begin date for the task to run.
        ------------------     --------------------------------------------------------------------
        start_date             Optional Datetime. The end date for the task to run.
        ------------------     --------------------------------------------------------------------
        title                  Optional String. The title of the scheduled task.
        ------------------     --------------------------------------------------------------------
        parameters             Optional Dict. Optional collection of Key/Values that will be given
                               to the notebook.
        ==================     ====================================================================

        :returns: Task

        """
        SPECIALS = {"reboot":   '@reboot',
                    "hourly":   '0 * * * *',
                    "daily":    '0 0 * * *',
                    "weekly":   '0 0 * * 0',
                    "monthly":  '0 0 1 * *',
                    "yearly":   '0 0 1 1 *',
                    "annually": '0 0 1 1 *',
                    "midnight": '0 0 * * *'}
        url = f"{self._url}/createTask"
        params = {
            "f" : "json",
            "title" : title,
            "type" : task_type,
            "parameters" : None,
            "itemId" : None,
            "startDate" : start_date,
            "endDate" : end_date,
            "minute" : None,
            "hour" : None,
            "dayOfMonth" : None,
            "month" : None,
            "dayOfWeek" : None,
            "maxOccurrences" : occurences
        }
        if isinstance(item, Item):
            params['itemId'] = item.itemid
        else:
            params['itemId'] = item
        if title is None:
            params.pop('title', None)
        if start_date:
            params['startDate'] = local_time_to_online(dt=start_date)
        else:
            params.pop('startDate', None)
        if end_date:
            params['endDate']= local_time_to_online(dt=end_date)
        else:
            params.pop('endDate', None)
        if cron in SPECIALS:
            cron = SPECIALS[cron]
        if cron:
            cron = cron.split(' ')
            while len(cron) < 5:
                cron.append("*")
            params.update({
                "minute" : cron[0],
                "hour" : cron[1],
                "dayOfMonth" : cron[2],
                "month" : cron[3],
                "dayOfWeek" : cron[4]
            })
        if parameters:
            params['parameters'] = json.dumps(parameters)
        res = self._gis._con.post(url, params)

        if 'success' in res and res['success']:
            url = f"{self._url}/{res['taskId']}"
            return Task(url, gis=self._gis)
        return res
    #----------------------------------------------------------------------
    @property
    def all(self) -> list:
        """returns all the current user's tasks"""
        if self._tasks is None:
            self._tasks = []
            url = f"{self._gis._portal.resturl}community/users/{self._user.username}/tasks"
            params = {
                'num' : 100,
                'start' : 1,
            }
            res = self._gis._con.get(url, params)
            for t in res['tasks']:
                url = f"{self._url}/{t['id']}"
                self._tasks.append(Task(url=url, gis=self._gis))
            while res['nextStart'] != -1:
                params['start'] = res['nextStart']
                res = self._gis._con.get(url, params)
                for t in res['tasks']:
                    url = f"{self._url}/{t['id']}"
                    self._tasks.append(Task(url=url, gis=self._gis))
        return self._tasks
    #----------------------------------------------------------------------
    @property
    def count(self) -> int:
        """
        Returns the number of tasks a user has

        :returns: Int
        """
        url = f"{self._gis._portal.resturl}community/users/{self._user.username}/tasks"
        params = {
            'num' : 1,
            'start' : 1,
        }
        res = self._gis._con.get(url, params)
        return res['total']






import uuid
import arcgis
from .._utils._basewidget import _BaseWidget
from .._utils._basewidget import NoDataProperties


class Indicator(_BaseWidget):
    """
    Creates a dashboard Indicator widget.

    =========================   ===========================================
    **Argument**                **Description**
    -------------------------   -------------------------------------------
    item                        Required Portal Item object. Item object can
                                be a Feature Layer or a MapWidget.
    -------------------------   -------------------------------------------
    name                        Optional string. Name of the Indicator
                                widget.
    -------------------------   -------------------------------------------
    title                       Optional string. Title or Caption for the
                                widget.
    -------------------------   -------------------------------------------
    layer                       Optional integer. Layer number when item is
                                a mapwidget.
    -------------------------   -------------------------------------------
    description                 Optional string. Description for the widget.
    =========================   ===========================================
    """
    def __init__(self, item, name='Indicator', layer=0, title='', description=''):
        super().__init__(name, title, description)

        if item.type not in ['Feature Service', 'mapWidget']:
            raise Exception("Please specify an item")

        self.item = item
        self.type = "indicatorWidget"
        self.layer = layer

        self._nodata = NoDataProperties._nodata_init()
        self._novalue = NoDataProperties._nodata_init()
        
        self._data = IndicatorData._create_data()
        self._reference = ReferenceData._create_data()

        self._max_display_features = 50
        self._show_last_update = True

    @classmethod
    def _from_json(cls, widget_json):
        gis = arcgis.env.active_gis
        itemid = widget_json["datasets"]["datasource"]["itemid"]
        name = widget_json["name"]
        item = gis.content.get(itemid)
        title = widget_json["caption"]
        description = widget_json["description"]
        indicator = Indicator(item, name, title, description)

        return indicator

    @property
    def data(self):
        """
        :return: Indicator Data object. Set data properties, categories and values.
        """
        return self._data
    
    @property
    def reference(self):
        """
        :return: Indicator reference object. Set data properties, categories and values.
        """
        return self._reference

    @property
    def max_features(self):
        """
        :return: max features for widget.
        """
        return self._max_features

    @max_features.setter
    def max_features(self, value):
        """
        Set max features for widget.
        """
        self._max_features = int(value)

    @property
    def show_last_update(self):
        """
        :return: show last update or not.
        """
        return self._show_last_update
    
    @show_last_update.setter
    def show_last_update(self, value):
        """
        Set show last update or not.
        """
        self.show_last_update = value
        if not isinstance(value, bool):
            self._show_last_update = True

    def _convert_to_json(self):

        self._reference_statistic = []
        self._statistic_defintion = []

        if self._data.value_type == "statistic":
            self._statistic_defintion.append({
                            "onStatisticField":self.data.value_field,
                            "outStatisticFieldName":"value",
                            "statisticType":self._data.statistic
                        })

        if self.item.type == 'mapWidget':
            wlayer = self.item.layers[self.layer]
            widget_id = self.item._id
            layer_id = wlayer["id"]
            self._datasource = {"id":str(widget_id)+'#'+str(layer_id)}
        else:
            self._datasource = {
                        "type": "featureServiceDataSource",
                        "itemId": self.item.itemid,
                        "layerId": 0,
                        "table": True
                    }

        self._datasets =[{
                    "type": "serviceDataset",
                    "dataSource": self._datasource,
                    "outFields": ["*"],
                    "groupByFields": [],
                    "orderByFields": [],
                    "statisticDefinitions": self._statistic_defintion,
                    "querySpatialRelationship": "esriSpatialRelIntersects",
                    "returnGeometry": False,
                    "clientSideStatistics": False,
                    "name": "main"
                }]
        if self._reference.reference_type == "fixed":
            self._datasets.append({
                    "type":"staticDataset",
                    "data":self._reference.fixed_value,
                    "name":"reference"
                })
        elif self._reference.reference_type == "statistic":
            self._datasets.append({
                    "type": "serviceDataset",
                    "dataSource": {
                        "type": "featureServiceDataSource",
                        "itemId": self.item.itemid,
                        "layerId": 0,
                        "table": True
                    },
                    "outFields": ["*"],
                    "groupByFields": [],
                    "orderByFields": [],
                    "statisticDefinitions": [{
                            "onStatisticField":self.reference.reference_field,
                            "outStatisticFieldName":"value",
                            "statisticType":self._reference.statistic
                        }],
                    "querySpatialRelationship": "esriSpatialRelIntersects",
                    "returnGeometry": False,
                    "clientSideStatistics": False,
                    "name": "reference"
                })

        json_data = {
            "type": "indicatorWidget",
            "defaultSettings": {
                "topSection": {"fontSize": 80, "textInfo": {}},
                "middleSection": {"fontSize": 160, "textInfo": {"text": "{value}"}},
                "bottomSection": {"fontSize": 80, "textInfo": {}}},
            "comparison": self._reference.reference_type,
            "valueField": self._data.value_field,
            "referenceField": self._reference.reference_field,
            "valueConversion": {"factor": self._data.factor, "offset": self._data.offset},
            "referenceConversion": {"factor": self._reference.factor, "offset": self._reference.offset},
            "valueFormat": {"name": "value", "type": "decimal", "prefix": False},
            "percentageFormat": {"name": "percentage", "type": "decimal", "prefix": False},
            "ratioFormat": {"name": "ratio", "type": "decimal", "prefix": False},
            "valueType": self._data.value_type,
            "noValueVerticalAlignment": self._novalue.alignment,
            "showCaptionWhenNoValue": self._novalue.show_title,
            "showDescriptionWhenNoValue": self._novalue.show_description,
            "datasets": self._datasets,
            "id": self._id,
            "name": self.name,
            "caption": self.title,
            "description": self.description,
            "showLastUpdate": self.show_last_update,
            "noDataVerticalAlignment": self._nodata.alignment,
            "showCaptionWhenNoData": self._nodata.show_title,
            "showDescriptionWhenNoData": self._nodata._show_description
        }
        if self._text_color:
            json_data['defaultSettings']['textColor'] = self._text_color

        if self._background_color:
            json_data['defaultSettings']['backgroundColor'] = self._background_color

        return json_data

class IndicatorData(object):

    @classmethod
    def _create_data(cls, value_type="statistic"):
        data = IndicatorData()

        data._value_type = "statistic"
        data._statistic = "count"
        data._value_field = "FID"
        data._factor = 1
        data._offset = 0
        data._filters = []

        data.value_type = value_type

        return data

    @property
    def value_type(self):
        """
        :return: value type of data.
        """
        return self._value_type

    @value_type.setter
    def value_type(self, value):
        """
        Set value type of data.
        """
        if value in ["statistic", "feature"]:
            self._value_type = str(value)
        else:
            raise Exception("Please set correct value type. Supported value types are 'statistic', 'feature'")

    @property
    def statistic(self):
        """
        :return: statistic for data.
        """
        return self._statistic

    @statistic.setter
    def statistic(self, value):
        """
        Set statistic for data.
        """
        if self._value_type == "statistic":
            if value in ['count', 'avg', 'min', 'max', 'stddev', 'sum']:
                self._statistic = value
        else:
            raise Exception("Can set statistic only for reference type 'statistic'")

    @property
    def factor(self):
        """
        :return: value conversion factor.
        """
        return self._factor

    @factor.setter
    def factor(self, value):
        """
        Set value conversion factor.
        """
        self._factor = value

    @property
    def offset(self):
        """
        :return: value conversion offset.
        """
        return self._offset

    @offset.setter
    def offset(self, value):
        """
        Set value conversion offset.
        """
        self._offset = value

    @property
    def value_field(self):
        """
        :return: value field for data.
        """
        return self._value_field

    @value_field.setter
    def value_field(self, value):
        """
        Set value field for data.
        """
        if self._value_type in ["feature", "statistic"]:
            self._value_field = value
        else:
            raise Exception("Can add value field only for value type 'feature' or 'statistic'")
        
    @property
    def filters(self):
        """
        :return: filters associated with widget
        """
        return self._filters
    
    def add_filter(self, field, join, condition, **kwargs):
        """
        Add filters associated with widget.
        """
        self._filter_field = field
        if join in ["AND", "OR"]:
            self._filter_join = join
        else:
            raise Exception("Please select from 'AND', 'OR'")
        if condition in ["between", "not between", "equal", "not equal", "greater than", "greater than or equal", "less than", "less than or equal", "is null", "is not null"]:
            self._filter_condition = condition
        else:
            raise Exception("Please select the right condition")

        if condition in ["between", "not between"]:
            self._val1 = kwargs.get('start')
            self,_val2 = kwargs.get('end')
            self._filters.append({"filtertype":self._filter_join, "field":self._filter_field, "operator":self._filter_condition, "start":self._val1, "end":self._val2})
        else:
            raise Exception("Please provide 'start' and 'end' values as parameters")

        if condition in ["equal", "not equal", "greater than", "greater than or equal", "less than", "less than or equal"]:
            self._val = kwargs.get('value')
            self._filters.append({"filtertype":self._filter_join, "field":self._filter_field, "operator":self._filter_condition, "value":self._val})
        else:
            raise Exception("Please provide a 'value' parameter for comparison")

class ReferenceData(object):

    @classmethod
    def _create_data(cls, reference_type="none"):
        data = ReferenceData()

        data._reference_type = "none"

        data._statistic = "count"
        data._reference_field = "FID"
        data._factor = 1
        data._offset = 0
        data._fixed_value = 0

        data.reference_type = reference_type

        return data

    @property
    def reference_type(self):
        """
        :return: reference type of reference.
        """
        return self._reference_type

    @reference_type.setter
    def reference_type(self, value):
        """
        Set reference type of reference.
        """
        if value in ["statistic", "feature", "none", "previous", "fixed"]:
            self._reference_type = str(value)
        else:
            raise Exception("Please set correct reference type. Supported reference types are 'none', 'statistic', 'previous', 'fixed', 'feature'")

    @property
    def statistic(self):
        """
        :return: statistic for reference.
        """
        return self._statistic

    @statistic.setter
    def statistic(self, value):
        """
        Set statistic for reference.
        """
        if self._value_type == "statistic":
            if value in ['count', 'avg', 'min', 'max', 'stddev', 'sum']:
                self._statistic = value
        else:
            raise Exception("Can set statistic only for reference type 'statistic'")

    @property
    def factor(self):
        """
        :return: value conversion factor.
        """
        return self._factor

    @factor.setter
    def factor(self, value):
        """
        Set value conversion factor.
        """
        self._factor = value

    @property
    def offset(self):
        """
        :return: value conversion offset.
        """
        return self._offset

    @offset.setter
    def offset(self, value):
        """
        Set value conversion offset.
        """
        self._offset = value

    @property
    def reference_field(self):
        """
        :return: reference field name.
        """
        return self._reference_field

    @reference_field.setter
    def reference_field(self, value):
        """
        Set reference field name.
        """
        if self._reference_type in ["feature", "previous"]:
            self._reference_field = value
        else:
            raise Exception("Can add reference field only for value type 'feature' or 'previous'")

    @property
    def fixed_value(self):
        """
        :return: fixed value of reference.
        """
        return self._fixed_value

    @fixed_value.setter
    def fixed_value(self, value):
        """
        :return: fixed value of reference.
        """
        if self._reference_type == "fixed":
            self._fixed_value = value
        else:
            raise Exception("Can set fixed value only for reference type 'fixed'")




import tensorflow as tf
from tensorflow.keras.layers import Input, Conv2D, Dropout, ReLU, BatchNormalization, \
                                    UpSampling2D, Reshape, Layer, AveragePooling2D, \
                                    MaxPool2D, GlobalAveragePooling2D, GlobalMaxPool2D, \
                                    Concatenate, Flatten, Dense, ZeroPadding2D, Cropping2D, SeparableConv2D
from tensorflow.keras import Model
from .._utils.common_tf import get_padding_info, get_cropping_info, get_skip_connection_info, get_channel_axis, UpSample2DToSize

def get_unet_tf_head_output(
    backbone_initalized, 
    data_format, 
    n_output_classes,
    skip_lyr_idxs, 
    skip_lyr_channels,
    mobile_optimized=False
    ):
    
    # Get channel axis
    ca = channel_axis = get_channel_axis(data_format=data_format)
    
    # Mobile Optimized convolutions
    bottle_bottleneck_ratio = 2
    bottle_out_ratio = 1
    if mobile_optimized:
        conv_obj = SeparableConv2D
        bottle_bottleneck_ratio = .7
        bottle_out_ratio = .5
    else:
        conv_obj = Conv2D

    ### Unet Middle Block - Bottleneck ###
    # conv block 1
    bottle_channels = backbone_initalized.output.shape[ca]
    bottle_block_conv1_bn0 = BatchNormalization(axis=channel_axis, name=f"bottle_block_conv1_bn0")(backbone_initalized.output)
    bottle_block_conv1 = conv_obj(round(bottle_channels*bottle_bottleneck_ratio), 3, data_format=data_format, padding='same', activation='relu', name=f"bottle_block_conv1")(bottle_block_conv1_bn0)
    bottle_block_conv1_bn1 = BatchNormalization(axis=channel_axis, name=f"bottle_block_conv1_bn1")(bottle_block_conv1)

    # conv block 3
    bottle_block_conv3 = conv_obj(round(bottle_channels*bottle_out_ratio), 3, data_format=data_format, padding='same', activation='relu', name=f"bottle_block_conv3")(bottle_block_conv1_bn1)#(bottle_block_conv2_drop1)
    bottle_block_conv3_bn1 = BatchNormalization(axis=channel_axis, name=f"bottle_block_conv3_bn1")(bottle_block_conv3)
    bottle_block_conv3_drop1 = Dropout(0.3, name="bottle_block_conv3_drop1")(bottle_block_conv3_bn1)
    
    block_output_layer = bottle_block_conv3_drop1

    ### Decoder Block ###
    for i, idx in enumerate(skip_lyr_idxs):
        # decoder block i

        skip_con_output = backbone_initalized.layers[idx].output
        block_in_channels = block_output_layer.shape[ca]
        conv1_channels = round(block_in_channels/2)
        conv2_channels = conv3_channels = round((block_in_channels+skip_lyr_channels[i])/2)
        if idx == skip_lyr_idxs[-1]:
            conv_obj = Conv2D
            conv1_channels = block_in_channels
            if not mobile_optimized:
                conv2_channels = round(conv2_channels/2)
        else:
            if mobile_optimized:
                conv2_channels = round((conv1_channels+skip_lyr_channels[i])/2)
        skip_con_size = skip_con_output.shape[2]

        
        # make input channels equal to target
        decoder_block_conv1 = SeparableConv2D(conv1_channels, 3, data_format=data_format, padding='same', activation='relu', name=f"decoder_block{i}_conv1")(block_output_layer)
        decoder_block_bn1 = BatchNormalization(axis=channel_axis, name=f"decoder_block{i}_bn1")(decoder_block_conv1)

        # Resize input
        decoder_block_upsample1 = UpSample2DToSize(skip_con_size, name=f"decoder_block{i}_resize1")(decoder_block_bn1)
        
        # Concatenate data from skip connection
        decoder_block_concat1 = Concatenate(axis=channel_axis, name=f"decoder_block{i}_concat1")([decoder_block_upsample1, skip_con_output])
        
        # first conv on concatenated data
        decoder_block_conv2 = conv_obj(conv2_channels, 3, data_format=data_format, padding='same', activation='relu', name=f"decoder_block{i}_conv2")(decoder_block_concat1)
        decoder_block_bn2 = BatchNormalization(axis=channel_axis, name=f"decoder_block{i}_bn2")(decoder_block_conv2)

        if idx == skip_lyr_idxs[-1]:
            block_output_layer = decoder_block_bn2
        else:
            # second conv on concatenated data
            decoder_block_conv3 = conv_obj(conv3_channels, 3, data_format=data_format, padding='same', activation='relu', name=f"decoder_block{i}_conv3")(decoder_block_bn2)
            decoder_block_bn3 = BatchNormalization(axis=channel_axis, name=f"decoder_block{i}_bn3")(decoder_block_conv3)
            block_output_layer = decoder_block_bn3        
        #
    decoder_block_output = block_output_layer

    # Final Classification Conv
    class_block_drop1 = Dropout(0.3, name="class_block_drop1")(decoder_block_output)
    class_block_conv1 = Conv2D(n_output_classes, 3, data_format=data_format, padding='same', activation=None, name=f"class_block_conv1")(class_block_drop1)


    # Cropping Block ##
    top_crop, bottom_crop, left_crop, right_crop = get_cropping_info( decoder_block_output, backbone_initalized.input, data_format=data_format)
    crop_block_crop1 = Cropping2D(
        cropping=((top_crop, bottom_crop), (left_crop, right_crop)), 
        data_format=data_format, name=f"crop_block_crop1"
    )(class_block_conv1)

    model_output_layer = crop_block_crop1
    
    return model_output_layer

def get_unet_tf_model(
    backbone_initalized, 
    data,
    mobile_optimized=False
    ):
    """
    creates a U-net model from the supplied backbone
    that can be used with the supplied data.
    """
    # Get Data Format
    data_format = tf.keras.backend.image_data_format() #'channels_first' or 'channels_last'

    # get skip connection information
    skip_lyr_idxs, skip_lyr_channels = get_skip_connection_info(backbone_initalized, data_format=data_format)

    # get head output
    head_output_layer = get_unet_tf_head_output(
        backbone_initalized=backbone_initalized, 
        data_format=data_format, 
        n_output_classes=data.c,
        skip_lyr_idxs=skip_lyr_idxs, 
        skip_lyr_channels=skip_lyr_channels,
        mobile_optimized=mobile_optimized
    )
    
    # Return Model
    return Model(inputs=backbone_initalized.input, outputs=head_output_layer)

def analyze_pred_unet_tf(activations):
    return tf.argmax(activations, axis=get_channel_axis())


import logging
log = logging.getLogger()

from arcgis.widgets._mapview._raster.engine import RasterData, MatplotlibRasterEngine

class RasterIOEngine(MatplotlibRasterEngine):
    def try_get_raster_data(self, image, bands) -> RasterData:
        try:
            return self._try_get_raster_data(image, bands)
        except Exception as e:
            return super().try_get_raster_data(image, bands)

    def _try_get_raster_data(self, image, bands) -> RasterData:
        import rasterio
        with rasterio.open(image) as dataset:
            extent = { 'xmin': dataset.bounds.left,
                       'ymin': dataset.bounds.bottom,
                       'xmax': dataset.bounds.right,
                       'ymax': dataset.bounds.top,
                       'spatialReference': {}}
            if dataset.crs and dataset.crs.to_epsg():
                extent['spatialReference']['wkid'] = dataset.crs.to_epsg()
            elif dataset.crs and dataset.crs.to_wkt():
                extent['spatialReference']['wkt'] = dataset.crs.to_wkt()
            else:
                extent = {}
            img_data = dataset.read()
            img_data = img_data.transpose((1,2,0)).squeeze()
            if bands:
                img_data = img_data[:,:, bands].squeeze()
            return RasterData(img_data = img_data,
                              extent = extent,
                              id_=str(hash(image)))



import uuid

class _BaseWidget(object):

    def __init__(self, name, title, description):
        
        self._title = ""
        self._name = ""
        self._description = ""
        self._background_color = None #"#ffffff"
        self._text_color = None #"#000000"
        self._id = str(uuid.uuid4())
        self._type = None

        self.title = title
        self.name = name
        self.description = description

        self._width = 1
        self._height = 1
    
    def _repr_html_(self):
        from arcgis.apps.dashboard import Dashboard
        url = Dashboard._publish_random(self)
        return f"""<iframe src={url} width=900 height=300>"""

    @property
    def type(self):
        """
        :return: widget type.
        """
        return self._type

    @type.setter
    def type(self, value):
        """
        Set widget type.
        """
        self._type = str(value)

    @property
    def name(self):
        """
        :return: widget name.
        """
        return self._name

    @name.setter
    def name(self, value):
        """
        Set widget name.
        """
        self._name = value

    @property
    def title(self):
        """
        :return: widget title.
        """
        return self._title

    @title.setter
    def title(self, value):
        """
        Set widget title.
        """
        self._title = value

    @property
    def description(self):
        """
        :return: widget description.
        """
        return self._description

    @description.setter
    def description(self, value):
        """
        Set widget description.
        """
        self._description = value
    
    @property
    def text_color(self):
        """
        :return: widget text color.
        """
        return self._text_color

    @text_color.setter
    def text_color(self, value):
        """
        Set widget text color.
        """
        if isinstance(value, str):
            self._text_color = value

    @property
    def background_color(self):
        """
        :return: widget background color.
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        """
        Set widget background color.
        """
        if isinstance(value, str):
            self._text_color = value

    @property
    def height(self):
        """
        :return: Height of the widget
        """
        return self._height

    @height.setter
    def height(self, value):
        """
        Set height of the widget, between 0 and 1.
        """
        if value > 1:
            self._height = 1
        elif value < 0:
            self._height = 0
        else:
            self._height = value

    @property
    def width(self):
        """
        :return: Width of the widget
        """
        return self._width

    @width.setter
    def width(self, value):
        """
        Set width of the widget, between 0 and 1.
        """
        if value > 1:
            self._width = 1
        elif value < 0:
            self._width = 0
        else:
            self._width = value
    
class Legend(object):

    @classmethod
    def _create_legend(cls, visible=True, placement="bottom"):
        legend = Legend()

        if placement not in ["bottom", "side"]:
            raise Exception("Please specify correct placement. Supported placement are 'bottom' and 'side'")

        legend._visibility = True
        legend._placement = "bottom"

        legend.visibility = visible
        legend.placement = placement

        return legend

    @property
    def visibility(self):
        """
        :return: Legend Visible or not.
        """
        return self._visibility

    @visibility.setter
    def visibility(self, value):
        """
        Set Legends visbility.
        """
        self._visibility = bool(value)

    @property
    def placement(self):
        """
        :return: Legend placement from 'bottom', 'side'
        """
        return self._placement

    @placement.setter
    def placement(self, value):
        """
        Set Legend placement from 'bottom', 'side'.
        """
        if value in ["bottom", "side"]:
            self._placement = value


class NoDataProperties(object):

    @classmethod
    def _nodata_init(cls, text="No Data", alignment="middle", show_title=True, show_description=True):

        nodata = NoDataProperties()

        if alignment not in ["top", "middle", "bottom"]:
            raise Exception("Please specify correct alignment. Supported alignment are 'top', 'bottom' and 'middle'")

        nodata._text = text
        nodata._alignment = alignment
        nodata._show_title = show_title
        nodata._show_description = show_description

        nodata.text = text
        nodata.alignment = alignment
        nodata.show_title = show_title
        nodata.show_description = show_description

        return nodata

    @property
    def text(self):
        """
        :return: No Data text.
        """
        return self._text
    
    @text.setter
    def text(self, value):
        """
        Set No Data text.
        """
        self._text = str(value)

    @property
    def alignment(self):
        """
        :return: No Data text vertical alignment.
        """
        return self._alignment
    
    @alignment.setter
    def alignment(self, value):
        """
        Set No Data text vertical alignment.
        """
        if value in ["top", "middle", "bottom"]:
            self._alignment = str(value)
    
    @property
    def show_title(self):
        """
        :return: No Data show title.
        """
        return self._show_title
    
    @show_title.setter
    def show_title(self, value):
        """
        Set No Data show title.
        """
        self._show_title = bool(value)

    @property
    def show_description(self):
        """
        :return: No Data show description.
        """
        return self._show_description

    @show_description.setter
    def show_description(self, value):
        """
        Set No Data show description.
        """
        self._show_description = bool(value)


def _auto_calculate_width(elements):
    import arcgis
    available_width = 1
    remaining_elements = len(elements)

    for el in elements:
        element_width = getattr(el, 'width', 1)
        if isinstance(el, dict) and not isinstance(el, arcgis.mapping.WebMap):
            element_width = el.get('width', 1)

        if element_width != 1:
            remaining_elements = remaining_elements - 1
            available_width = available_width - element_width

    if remaining_elements > 0:
        available_width = float(available_width / remaining_elements)

    for el in elements:
        if isinstance(el, dict) and not isinstance(el, arcgis.mapping.WebMap):
            if el.get('width', 1) == 1:
                el['width'] = available_width
        elif getattr(el, 'width', 1) == 1:
            el.width = available_width

    return elements


def _auto_calculate_height(elements):
    import arcgis
    available_height = 1
    remaining_elements = len(elements)

    for el in elements:
        element_height = getattr(el, 'height', 1)
        if isinstance(el, dict) and not isinstance(el, arcgis.mapping.WebMap):
            element_height = el.get('height', 1)

        if element_height != 1:
            remaining_elements = remaining_elements - 1
            available_height = available_height - element_height

    if remaining_elements > 0:
        available_height = float(available_height / remaining_elements)

    for el in elements:
        if isinstance(el, dict) and not isinstance(el, arcgis.mapping.WebMap):
            if el.get('height', 1) == 1:
                el['height'] = available_height
        elif getattr(el, 'height', 1) == 1:
            el.height = available_height

    return elements


def add_row(elements, height=1):
    """
    Creates a Row Layout.

    =========================   ===========================================
    **Argument**                **Description**
    -------------------------   -------------------------------------------
    elements                    Required List. Widgets that have to be part
                                of the row.
    -------------------------   -------------------------------------------
    height                      Optional int. Height of the row.
    =========================   ===========================================
    """

    elements = _auto_calculate_width(elements)
    json = {
        "type": "stackLayoutElement",
        "orientation": "col",
        "elements": [],
        "width": 1,
        "height": height,
        'widgets': []
    }

    for el in elements:
        if not hasattr(el, '_id'):
            json['widgets'] = json['widgets'] + el['widgets']
            del el['widgets']
            json['elements'].append(el)
        else:
            json['elements'].append({
                'type': 'itemLayoutElement',
                'id': el._id,
                'height': el.height,
                'width': el.width
            })
            json['widgets'].append(el)

    return json


def add_column(elements, width=1):
    """
    Creates a Column Layout.

    =========================   ===========================================
    **Argument**                **Description**
    -------------------------   -------------------------------------------
    elements                    Required List. Widgets that have to be part
                                of the row.
    -------------------------   -------------------------------------------
    width                       Optional int. Width of the Column.
    =========================   ===========================================
    """

    elements = _auto_calculate_height(elements)
    json = {
        "type": "stackLayoutElement",
        "orientation": "row",
        "elements": [],
        "width": width,
        "height": 1,
        'widgets': []
    }

    for el in elements:
        if not hasattr(el, '_id'):
            json['widgets'] = json['widgets'] + el['widgets']
            del el['widgets']
            json['elements'].append(el)
        else:
            json['elements'].append({
                'type': 'itemLayoutElement',
                'id': el._id,
                'height': el.height,
                'width': el.width
            })
            json['widgets'].append(el)

    return json

